/*
  Copyright 2007 Eurotech S.P.A. (Luca Calligaris)
  Copyright 2007 Christian Pellegrin <chripell@gmail.com> EVOL S.R.L.
  
  This software is released under the GPL v2. Please see the file
  COPYING in the top-level directory.
*******************************************************************************/

#include "AP5200.h"
#include "I2C.h"

#include "utils.h"

/*
The following file contains AP5200
table for PAL/NTSC, QUAD, FULL...
for various operations
*/
#include "AP5200_T.c"

/*OSD font table */
#include "font.c"

#define	COMMAND		0x2000
#define	STOP		0x2004
#define	DATA		0x2008

#define	COMMAND_MOT_WRITE		0x01
#define	COMMAND_MOT_READ		0x02
#define	COMMAND_FONT_WRITE	COMMAND_MOT_WRITE
#define	COMMAND_IMG_READ		0x04
#define	COMMAND_IMG_WRITE		0x05



const AP5200_COLOR WHITE	= {0xEB,0xEB,0xEB};
const AP5200_COLOR YELLOW 	= {0xD2,0x10,0x92};
const AP5200_COLOR CYAN	  	= {0xAA,0xA6,0x10};
const AP5200_COLOR GREEN  	= {0x91,0x36,0x22};
const AP5200_COLOR MAGENTA	= {0x6A,0xCA,0xDE};
const AP5200_COLOR RED		= {0x51,0x5A,0xF0};
const AP5200_COLOR BLUE		= {0x29,0xF0,0x6E};
const AP5200_COLOR BLACK 	= {0x10,0x80,0x80};


/*Low level function*/			
///////////////////////////////////////////////////////////////////////////////
// AP5200 Image Processor interface
unsigned short	AP5200_ImgRegReadSha( PETHJPG2K pDevice, unsigned char Address )
{
	/*Most of AP5200 image processor registers are write only so we'll return our shadow copy*/
	return pDevice->AP5200_Shadow[Address];
}

void AP5200_ImgRegWriteSingle( PETHJPG2K pDevice, unsigned char bSlave, unsigned char Address, unsigned short Data )
{

		//stop previous cycle if any is active
		WD(pDevice->PLX9056_BAR2[bSlave]+STOP,0);
	
		//Cycle: W(Addr)-W(Lsb)-W(Msb)
	
		WD(pDevice->PLX9056_BAR2[bSlave]+COMMAND, COMMAND_IMG_WRITE );
		WD(pDevice->PLX9056_BAR2[bSlave]+DATA, Address );
		WD(pDevice->PLX9056_BAR2[bSlave]+DATA, (unsigned char)Data );
		WD(pDevice->PLX9056_BAR2[bSlave]+DATA, (unsigned char)(Data>>8) );
		WD(pDevice->PLX9056_BAR2[bSlave]+STOP,0);
}

void AP5200_ImgRegWrite( PETHJPG2K pDevice, unsigned char Address, unsigned short Data )
{
	UCHAR bSlave = 0;
	
	for ( ; bSlave < pDevice->Bridges; bSlave++ )
		AP5200_ImgRegWriteSingle(pDevice,bSlave,Address,Data);
	/*Save shadow*/
	pDevice->AP5200_Shadow[Address] = Data;
}

///////////////////////////////////////////////////////////////////////////////
//AP5200 Motion Processor interface 
/*Channel registers*/
unsigned short	AP5200_MotRegRead( PETHJPG2K pDevice, unsigned char Address, unsigned char bSlave)
{
  int i;
  unsigned short Data = 0;
  for(i=0; i<2; i++) { //DELME
    Data = 0;

	//stop previous cycle if any is active
	WD(pDevice->PLX9056_BAR2[bSlave]+STOP,0);

	//Cycle: W(Addr)-W(0xFF)-R(Msb)-R(Lsb)

	WD(pDevice->PLX9056_BAR2[bSlave]+COMMAND, COMMAND_MOT_READ );
	WD(pDevice->PLX9056_BAR2[bSlave]+DATA, 0xFF );
	WD(pDevice->PLX9056_BAR2[bSlave]+DATA, Address );
	Data = (unsigned short)(RD(pDevice->PLX9056_BAR2[bSlave]+DATA) & 0xFFL);
	Data <<= 8;
	Data |=  (unsigned short)(RD(pDevice->PLX9056_BAR2[bSlave]+DATA) & 0xFFL);
	WD(pDevice->PLX9056_BAR2[bSlave]+STOP,0);
  }
	return Data;
}
unsigned short	AP5200_MotRegRead1( PETHJPG2K pDevice, unsigned char Address, unsigned char bSlave) //DELME
{
  int i;
  unsigned short Data = 0;
  for(i=0; i<2; i++) { //DELME
    Data = 0;

	//stop previous cycle if any is active
	WD(pDevice->PLX9056_BAR2[bSlave]+STOP,0);

	//Cycle: W(Addr)-W(0xFF)-R(Msb)-R(Lsb)

	WD(pDevice->PLX9056_BAR2[bSlave]+COMMAND, COMMAND_MOT_READ );
	WD(pDevice->PLX9056_BAR2[bSlave]+DATA, 0xFF );
	WD(pDevice->PLX9056_BAR2[bSlave]+DATA, Address );
	Data = (unsigned short)(RD(pDevice->PLX9056_BAR2[bSlave]+DATA) & 0xFFL);
	Data <<= 8;
	Data |=  (unsigned short)(RD(pDevice->PLX9056_BAR2[bSlave]+DATA) & 0xFFL);
	WD(pDevice->PLX9056_BAR2[bSlave]+STOP,0);
  }
	return Data;
}
/*Common registers*/
unsigned short	AP5200_MotRegReadSha( PETHJPG2K pDevice, unsigned char Address )
{
	/*Most of AP5200 motion processor common registers are write only so we'll return our shadow copy*/
	return pDevice->AP5200_Shadow[Address];	
}

void AP5200_MotRegWriteSingle( PETHJPG2K pDevice, unsigned char Address, unsigned short Data, UCHAR bSlave) {
  //stop previous cycle if any is active
  WD(pDevice->PLX9056_BAR2[bSlave]+STOP,0);
  
  //Cycle: W(Addr)-W(0xFF)-W(Msb)-W(Lsb)
  
  WD(pDevice->PLX9056_BAR2[bSlave]+COMMAND, COMMAND_MOT_WRITE );
  WD(pDevice->PLX9056_BAR2[bSlave]+DATA, 0xFF );
  WD(pDevice->PLX9056_BAR2[bSlave]+DATA, Address );
  WD(pDevice->PLX9056_BAR2[bSlave]+DATA, (unsigned char)((Data & 0xFF00)>>8) );
  WD(pDevice->PLX9056_BAR2[bSlave]+DATA, (unsigned char)(Data & 0xFF) );
  WD(pDevice->PLX9056_BAR2[bSlave]+STOP,0);
}


void AP5200_MotRegWrite( PETHJPG2K pDevice, unsigned char Address, unsigned short Data )
{
	UCHAR bSlave = 0;
	
	for ( ; bSlave < pDevice->Bridges; bSlave++ )
	{	
	  AP5200_MotRegWriteSingle(pDevice, Address, Data, bSlave);
	}
	/*Save shadow*/
	pDevice->AP5200_Shadow[Address] = Data;
}




///////////////////////////////////////////////////////////////////////////////
//Init AP5200 interface
unsigned char	AP5200_Init( PETHJPG2K pDevice, unsigned char bNtsc )
{
	unsigned char* pTable = bNtsc ? tNTSC_MAIN : tPAL_MAIN;
	//unsigned short Data;
	unsigned char i;
	

	//clr SDRAM
	AP5200_ImgRegWrite( pDevice,mixed_4,0x0001);
	Sleep(300);

	//Load video input tables
	for ( i = 0; i < (4*pDevice->Bridges); i++ )
		if (!SAA7113H_Init(pDevice,i,bNtsc))
			return 0;
	
	//Load init table
	AP5200_TableLoad( pDevice,pTable );	

    //Load font RAM
	AP5200_FontLoad(pDevice);
	
	//Enable and clear monitor path OSD 
	AP5200_OSDEnable( pDevice,MONITOR_PATH,TRUE, YELLOW, BLUE, RED );
	AP5200_OSDClear( pDevice,MONITOR_PATH);

	//Enable and clear master record path OSD
	AP5200_OSDEnable( pDevice,RECORD_MST_PATH,TRUE, YELLOW, BLUE, RED );
	AP5200_OSDClear( pDevice,RECORD_MST_PATH);
	//Enable and clear slave record path OSD
	AP5200_OSDEnable( pDevice,RECORD_SLV_PATH,TRUE, YELLOW, BLUE, RED );
	AP5200_OSDClear( pDevice,RECORD_SLV_PATH);


	//Erase SDRAM memory
	AP5200_ImgRegWrite( pDevice,mixed_4,0x0000);
	Sleep(21);

	AP5200_ImgRegWrite(pDevice, mon_window_fill,0xFFFF);
	AP5200_ImgRegWrite(pDevice, mon_window_fill,0x0);

	// disable motion detection
	AP5200_MotRegWrite(pDevice,MDCH_sel,0x0000);
	wait_sync();
	AP5200_MotRegWrite(pDevice,INT_mask,0xffff);
	AP5200_MotRegWrite(pDevice,MD_off,0xffff);
	AP5200_MotRegWrite(pDevice,bypass,0xffff);

	return 1;     
}
///////////////////////////////////////////////////////////////////////////////
//Set AP5200 interface
unsigned char 	AP5200_SetMode( PETHJPG2K pDevice, PAP5200_MODE_PARAMS  pParams)
{
	unsigned short	ChannelOn = 0, Scale = 0;
	unsigned char	bSlave,i,Rotate, MaxChannels = ( pDevice->Bridges == 2 ) ? 7 : 3;
	unsigned short	NTSCQuadChStart[4] = {0,0x59,0x3C59,0x3C00};
	unsigned short	PALQuadChStart[4] =	{0,0x59,0x4859,0x4800};
	unsigned short	NTSCAllChStart[8] = {0,0x3B,0x76,0x2876,0x5076,0x503B,0x5000,0x2800};	
	unsigned short	PALAllChStart[8] = {0,0x3B,0x76,0x3076,0x6076,0x603B,0x6000,0x3000};	
	unsigned char	Mode = pParams->Mode;
	unsigned short	rec_status_data, rec_priority_data, rec_ch_on_data;

	if ( Mode > MODE_REC_SWITCH )
		return 0;

	//clr SDRAM
	AP5200_ImgRegWrite( pDevice,mixed_4,0x0001);
	Sleep(300);

	switch( Mode )
	{
		case MODE_MON_QUAD:
			for ( i = 0; i < 4; i++ )
				if ( pParams->MON_QUAD.Channel[i] > MaxChannels )
					return 0;
					
			ChannelOn = 0;

			for ( i = 0; i < 4; i++ )
			{
				AP5200_ImgRegWrite(pDevice,ch1_start + pParams->MON_QUAD.Channel[i], 
					pDevice->Flags.bNtsc ? NTSCQuadChStart[i] : PALQuadChStart[i]);
				AP5200_ImgRegWrite(pDevice,ch1_length + pParams->MON_QUAD.Channel[i],
					pDevice->Flags.bNtsc ? NTSC_QUAD_DIM: PAL_QUAD_DIM);
				
				ChannelOn |= 0x01 << pParams->MON_QUAD.Channel[i];
			}
			
			/*Actually here we set every channel ratio to quarter screen...*/
			AP5200_ImgRegWrite( pDevice, MST_ch_scl_mode, 0x1111);
			AP5200_ImgRegWrite( pDevice, SLV1_ch_scl_mode, 0x1111);			
			
			/*Activate channels*/
			AP5200_ImgRegWrite(pDevice,mon_ch_on,ChannelOn);
		break;
		
		case MODE_MON_FULL:
			if ( pParams->MON_FULL.Channel > MaxChannels )
				return 0;
				
			/*Start in (0,0)*/
			AP5200_ImgRegWrite(pDevice,ch1_start + pParams->MON_FULL.Channel,0 );
			AP5200_ImgRegWrite(pDevice,ch1_length + pParams->MON_FULL.Channel,
				pDevice->Flags.bNtsc ? NTSC_FULL_DIM : PAL_FULL_DIM);
			
			/*Actually here we set every channel ratio to full screen...*/
			AP5200_ImgRegWrite( pDevice, MST_ch_scl_mode, 0);
			AP5200_ImgRegWrite( pDevice, SLV1_ch_scl_mode, 0);
			
			/*Activate channel*/
			AP5200_ImgRegWrite(pDevice,mon_ch_on, 1 << pParams->MON_FULL.Channel);
		break;
		
		case MODE_MON_ALL:
			for (  i = 0; i < MaxChannels+1; i++ )
				if ( pParams->MON_ALL.Channel[i] > MaxChannels )
					return 0;
			
			ChannelOn = 0;

			for ( i = 0; i < 8; i++ )
			{
				AP5200_ImgRegWrite(pDevice,ch1_start + pParams->MON_ALL.Channel[i], 
					pDevice->Flags.bNtsc ? NTSCAllChStart[i] : PALAllChStart[i]);
				AP5200_ImgRegWrite(pDevice,ch1_length + pParams->MON_ALL.Channel[i],
					pDevice->Flags.bNtsc ? NTSC_ALL_DIM: PAL_ALL_DIM);
				
				ChannelOn |= (0x01 << pParams->MON_QUAD.Channel[i]);
			}
			
			/*Actually here we set every channel ratio to ninth of screen...*/
			AP5200_ImgRegWrite( pDevice, MST_ch_scl_mode, 0x2222);
			AP5200_ImgRegWrite( pDevice, SLV1_ch_scl_mode, 0x2222);			
			
			/*Activate channels*/
			AP5200_ImgRegWrite(pDevice,mon_ch_on,ChannelOn);			
		break;
		
		case MODE_MON_CUSTOM:
			if ( pParams->MON_CUSTOM.Channel > MaxChannels )
				return 0;
				
			/*Display area must fit the screen*/
			if ( ( pParams->MON_CUSTOM.Width + pParams->MON_CUSTOM.StartX ) > ( pDevice->Flags.bNtsc ? NTSC_WIDTH : PAL_WIDTH ) )
				return 0;
			
			if ( ( pParams->MON_CUSTOM.Height + pParams->MON_CUSTOM.StartY ) > ( pDevice->Flags.bNtsc ? NTSC_HEIGHT : PAL_HEIGHT ) )
				return 0;
				
			/*Only first channel of each chip can scale to 2/3 x 2/3*/
			if ( (pParams->MON_CUSTOM.Ratio == SCALE_2_3_X_2_3) && ( pParams->MON_CUSTOM.Channel != 0 )
					&& ( pParams->MON_CUSTOM.Channel != 4 ) )
				return 0;
			
			if ( pParams->MON_CUSTOM.Ratio > SCALE_1_4_X_1_2 )
				return 0; 
				
			/*Program leaving other channels settings unchanged*/
			ChannelOn = pParams->MON_CUSTOM.StartY;
			ChannelOn <<= 8;
			ChannelOn += pParams->MON_CUSTOM.StartX;
			AP5200_ImgRegWrite(pDevice,ch1_start + pParams->MON_CUSTOM.Channel,ChannelOn); 
			
			ChannelOn = pParams->MON_CUSTOM.Height;
			ChannelOn <<= 8;
			ChannelOn += pParams->MON_CUSTOM.Width;			
			AP5200_ImgRegWrite(pDevice,ch1_length + pParams->MON_CUSTOM.Channel,ChannelOn); 
			
			Scale = AP5200_ImgRegReadSha(pDevice, (pParams->MON_CUSTOM.Channel < 4) ? MST_ch_scl_mode : SLV1_ch_scl_mode);
			Rotate = 4*(pParams->MON_CUSTOM.Channel%4);
			ChannelOn = pParams->MON_CUSTOM.Ratio;
			
#if 1
			Scale &= ~ (0xf << Rotate);
			Scale |= ChannelOn << Rotate;
#else			
			/*Mask*/
			__asm
			{
				push eax
				push ecx
				mov	ax, Scale
				mov cl, Rotate
				ror	ax, cl
				and ax, 0FFF0h
				or	ax, ChannelOn
				rol ax, cl
				mov Scale, ax
				pop	ecx	
				pop	eax
			}
#endif
			
			AP5200_ImgRegWrite(pDevice, (pParams->MON_CUSTOM.Channel < 4) ? MST_ch_scl_mode : SLV1_ch_scl_mode, Scale);
	
	
			
			ChannelOn = AP5200_ImgRegReadSha(pDevice, mon_ch_on);
			ChannelOn |= 0x01 << pParams->MON_CUSTOM.Channel;
			AP5200_ImgRegWrite(pDevice, mon_ch_on, ChannelOn);
		
		break;

		case MODE_REC_QUAD:
/*		
			bSlave = pParams->REC_QUAD.bSlave;
			AP5200_TableLoadSingle(pDevice,bSlave,pDevice->Flags.bNtsc ? tNTSC_REC_QUAD : tPAL_REC_QUAD);
																	
			AP5200_ImgRegWriteSingle(pDevice,bSlave,rec_status, (bSlave ? 0xF0 : 0x0F));
			AP5200_ImgRegWriteSingle(pDevice,bSlave,rec_priority, (bSlave ? 0xF0 : 0x0F));
			AP5200_ImgRegWriteSingle(pDevice,bSlave,rec_ch_on,(bSlave ? 0xF0 : 0x0F));
			AP5200_ImgRegWriteSingle(pDevice,bSlave,rec_main,0x8004);
*/
			bSlave = pParams->REC_QUAD.bSlave;

			//Save previous settings, don't alter the other record path
			rec_status_data = AP5200_ImgRegReadSha(pDevice,rec_status) | (bSlave ? 0xF0 : 0x0F);
			rec_priority_data = AP5200_ImgRegReadSha(pDevice,rec_priority) | (bSlave ? 0xF0 : 0x0F);
			rec_ch_on_data = AP5200_ImgRegReadSha(pDevice,rec_ch_on) | (bSlave ? 0xF0 : 0x0F);
			
			AP5200_ImgRegWrite(pDevice,rec_main,0x8004);

			//Load table
			AP5200_TableLoad(pDevice,pDevice->Flags.bNtsc ? tNTSC_REC_QUAD : tPAL_REC_QUAD);

			//New settings
			
			//AP5200_ImgRegWrite(pDevice,rec_status, rec_status_data);
			//AP5200_ImgRegWrite(pDevice,rec_priority, rec_priority_data);
			AP5200_ImgRegWrite(pDevice,rec_ch_on, rec_ch_on_data);

		break;
		
		case MODE_REC_FULL:
			if (pParams->REC_FULL.Channel > 3)
				bSlave = 1;
			else 
				bSlave = 0;

			rec_status_data = AP5200_ImgRegReadSha(pDevice,rec_status) & ~(bSlave ? 0xF0 : 0x0F);
			rec_priority_data = AP5200_ImgRegReadSha(pDevice,rec_priority) & ~(bSlave ? 0xF0 : 0x0F);
			rec_ch_on_data = AP5200_ImgRegReadSha(pDevice,rec_ch_on) & ~(bSlave ? 0xF0 : 0x0F);

			AP5200_ImgRegWrite(pDevice,rec_main,0x8012);	

			AP5200_TableLoad(pDevice,pDevice->Flags.bNtsc ? tNTSC_REC_FULL : tPAL_REC_FULL);

			AP5200_ImgRegWrite(pDevice,rec_status, rec_status_data | (0x01 << pParams->REC_FULL.Channel));
			//AP5200_ImgRegWrite(pDevice,rec_priority,rec_priority_data | (0x01 << pParams->REC_FULL.Channel));
			//AP5200_ImgRegWrite(pDevice,rec_ch_on,rec_ch_on_data | (0x01 << pParams->REC_FULL.Channel));

		break;
		
		case MODE_REC_SWITCH:
			
		break;

		default: return 0;	
	}
	
	
	AP5200_ImgRegWrite( pDevice,mixed_4,0x0000);
	
	return 1;
	
}
///////////////////////////////////////////////////////////////////////////////
//Enable OSD			
void AP5200_OSDEnable( PETHJPG2K pDevice, unsigned char TargetPath, unsigned char bEnable,
	  AP5200_COLOR OsdColorA, AP5200_COLOR OsdColorB, AP5200_COLOR OsdColorC)
{
	unsigned short ColorData;
	unsigned char  OsdCtrl = (TargetPath == MONITOR_PATH) ? mon_osd_ctrl : rec_osd_ctrl;


	if (TargetPath == MONITOR_PATH)
		AP5200_ImgRegWrite( pDevice, mon_osd_ctrl, 0x03 );
	else
	{
		ColorData = AP5200_ImgRegReadSha( pDevice, OsdCtrl );
		AP5200_ImgRegWrite( pDevice, OsdCtrl, ColorData | ((TargetPath == RECORD_SLV_PATH) ? 0x30 : 0x03) );
	}

	if ( !bEnable )
		return;

	//big mess...
	ColorData = (unsigned short)OsdColorA.Cb;
	ColorData <<=8;
	ColorData |= OsdColorA.Y;
	AP5200_ImgRegWrite( pDevice,0xA9,ColorData);

	ColorData = (unsigned short)OsdColorB.Y;
	ColorData <<=8;
	ColorData |= OsdColorA.Cr;
	AP5200_ImgRegWrite( pDevice,0xAA,ColorData);
     
	ColorData = (unsigned short)OsdColorB.Cr;
	ColorData <<=8;
	ColorData |= OsdColorB.Cb;
	AP5200_ImgRegWrite( pDevice,0xAB,ColorData);     
     
	ColorData = (unsigned short)OsdColorC.Cb;
	ColorData <<=8;
	ColorData |= OsdColorC.Y;     
	AP5200_ImgRegWrite( pDevice,0xAC,ColorData);     
     
	ColorData = (unsigned short)OsdColorC.Cr;
	AP5200_ImgRegWrite( pDevice,0xAD,ColorData);    
	

	if (TargetPath == MONITOR_PATH)
		AP5200_ImgRegWrite( pDevice, mon_osd_ctrl, 0x00 );
	else
	{
		ColorData = AP5200_ImgRegReadSha( pDevice, OsdCtrl );
		AP5200_ImgRegWrite( pDevice, OsdCtrl, ColorData | ((TargetPath == RECORD_SLV_PATH) ? ~0x30 : ~0x03) );
		
	}
     

   
}
///////////////////////////////////////////////////////////////////////////////
//Clear OSD
void AP5200_OSDClear( PETHJPG2K pDevice, unsigned char TargetPath )
{
	unsigned char  OsdCtrl = (TargetPath == MONITOR_PATH) ? mon_osd_ctrl : rec_osd_ctrl;
	unsigned short Data;

	if (TargetPath == MONITOR_PATH)
	{
		AP5200_ImgRegWrite( pDevice, OsdCtrl, 0x02 );
		AP5200_ImgRegWrite( pDevice, OsdCtrl, 0x00 );
	}
	else
	{
		Data = AP5200_ImgRegReadSha( pDevice, OsdCtrl );
		AP5200_ImgRegWrite( pDevice, OsdCtrl, Data | ((TargetPath == RECORD_SLV_PATH) ? 0x20 : 0x02) );
		AP5200_ImgRegWrite( pDevice, OsdCtrl, Data & ((TargetPath == RECORD_SLV_PATH) ? ~0x20 : ~0x02) );
	}
}
///////////////////////////////////////////////////////////////////////////////
//AP5200 translation ASCII->OSD RAM font
//If a charachter is not recognized  ( see font.c ) will be substituted by a blank space
unsigned char AP5200_ASCIIToFont(char ToConvert)
{
	  //numbers
	  if ( (ToConvert >= '0') && (ToConvert <= '9' ) )
			 return (S00 + (ToConvert-'0'));
	  //lowercase
	  else if ( (ToConvert >= 'a') && (ToConvert <= 'z' ) )
			 return (aaa + (ToConvert-'a'));
	  //uppercase
	  else if ( (ToConvert >= 'A') && (ToConvert <= 'Z' ) )
			 return (AAA + (ToConvert-'A'));
	  //special chars
	  else if(  ToConvert == '.' )
			 return DOT;
	  else if(  ToConvert == ':' )
			 return COL;
	  else
	  //blank
			 return BLK;
}
///////////////////////////////////////////////////////////////////////////////
//Write a string on OSD
//
//x=[0,OSD_COLS], y=[0,OSD_ROWS]
unsigned char AP5200_OSDWriteString( PETHJPG2K pDevice, unsigned char TargetPath,
					unsigned char x, unsigned char y, char *String, unsigned short Color )
{
	  unsigned short    StrLen = 0;
	  char*     ptr = String;
	  unsigned char     OsdAddr = (TargetPath == MONITOR_PATH) ? mon_osd_addr : 
									((TargetPath == RECORD_SLV_PATH) ? SLV1_rec_osd_addr : MST_rec_osd_addr);
	  unsigned char     OsdData = (TargetPath == MONITOR_PATH) ? mon_osd_data : 
									((TargetPath == RECORD_SLV_PATH) ? SLV1_rec_osd_addr : MST_rec_osd_addr);

	  //Calculate strlen
	  while(*ptr++) StrLen++;

	  //validate coordinates
	  if ( (x > (OSD_COLS-1)) || (y > (OSD_ROWS-1)) )
			 return 0;

	  //validate color
	  switch ( Color )
	  {
			 case OSD_COLOR_A:
			 case OSD_COLOR_B:
			 case OSD_COLOR_C:
					break;
			 default:
					Color = OSD_COLOR_DEF;
	  }

	  //Does string fits in one row?
	  if ( (x + StrLen) > OSD_COLS )
			 return 0;


	  for ( ; StrLen > 0; StrLen-- )
	  {
			 //Position
			 AP5200_ImgRegWrite( pDevice,OsdAddr,(y*OSD_COLS) + x++);
			 //Data
			 AP5200_ImgRegWrite( pDevice,OsdData,
					((unsigned short)AP5200_ASCIIToFont(*String++) | Color));
	  }


	  return 1;
}

/*--------------------------------------------------------------------------------------
	Function:		AP5200_Zoom ( PAP5200_RECT Zoom , PAP5200_RECT View, unsigned char ZoomCoeff, BOOL EnableZoom )
	Purpose:		Zoom specified output screen area
	Parameters:		
					Zoom:		portion of the screen to be zoomed
					View:		portion of the screen displaying zoomed area
					ZoomCoeff:	zoom coefficient ( see Notes )
					EnableZoom:	if FALSE zoom is disabled and any other parameter is ignored
								if TRUE zoom is enabled
	Return:			FALSE if invalid parameters are passed, TRUE otherwise  
	Notes:			AP5200 is able to perform a linear zoom from 2x through 4x.
					The relationship between zoom ratio and ZoomCoeff is ZoomRatio = 256/ZommCoeff:
					2x Zoom => ZoomCoeff == 0x80
					4x Zoom => ZoomCoeff == 0x40
					Any value > 0x80 or < 0x40 is invalid.
					Some useful zoom coefficients are #define'd in AP5200.H
					Parameters must satisfy the following inequations:

					(4*Zoom.w) * (256/ZoomCoeff) <= 720
					(2*Zoom.h) * (256/ZoomCoeff) <= 240
					(View.x+View.w) * 4 <= 720
					(View.y+View.h) * 2 <= 240
---------------------------------------------------------------------------------------*/
unsigned char AP5200_Zoom ( PETHJPG2K pDevice, PAP5200_RECT Zoom , PAP5200_RECT View, 
								unsigned char ZoomCoeff, unsigned char bEnableZoom )
{
     unsigned short Data;
     
     
     if ( !bEnableZoom )
     {
          //disable
          AP5200_ImgRegWrite(pDevice,zoom_ctrl,0x00);
          return TRUE;
     }
     
               
     //Validate parameters
     if ( !((ZoomCoeff >= ZOOM_4_0X ) && (ZoomCoeff <= ZOOM_2_0X)) )
          return FALSE;
          
     if( !(    (((Zoom->w) * (256/ZoomCoeff)) <= 720) &&
               ((Zoom->h) * (256/ZoomCoeff) <= 240)   ))
          return FALSE;

     if( !(    (((View->x+View->w)) <= 720) &&
               (((View->y+View->h)) <= 240)   ))
          return FALSE;      

     Data = 0;
     Data = (Zoom->y)/Y_DISPLAY_UNIT;
     Data <<=8;
     Data |= (Zoom->x)/X_DISPLAY_UNIT;
     
     AP5200_ImgRegWrite(pDevice,zoom_start,Data);

     
     Data = 0;
     Data = Zoom->h/Y_DISPLAY_UNIT;
     Data <<=8;
     Data |= Zoom->w/X_DISPLAY_UNIT;

     AP5200_ImgRegWrite(pDevice,zoom_length,Data);


     Data = 0;
     Data = View->y/Y_DISPLAY_UNIT;
     Data <<=8;
     Data |= View->x/X_DISPLAY_UNIT;
     
     AP5200_ImgRegWrite(pDevice,view_start,Data);


     Data = 0;
     Data = View->h/Y_DISPLAY_UNIT;
     Data <<=8;
     Data |= View->w/X_DISPLAY_UNIT;
     
     AP5200_ImgRegWrite(pDevice,view_length,Data);               
          
      
     Data = 0;
     Data = ZoomCoeff;                   
     Data <<= 8;
     Data |= 0x80;
     
     //enable
     AP5200_ImgRegWrite(pDevice,zoom_ctrl,Data);
     
     
     return 1;          

}
			
///////////////////////////////////////////////////////////////////////////////
//Erase AP5200 SDRAM memory 
void AP5200_EraseSdram(PETHJPG2K pDevice )
{
	AP5200_ImgRegWrite( pDevice,mixed_4,0x0001);
	Sleep(200);
	AP5200_ImgRegWrite( pDevice,mixed_4,0x0000);	
}
///////////////////////////////////////////////////////////////////////////////
//Load AP5200 registers with predefined table
void AP5200_TableLoad( PETHJPG2K pDevice, unsigned char *ptr)
{
	unsigned char addr, number;
	unsigned short Data;

	addr = *ptr;		//start address
	ptr++;
	number = *ptr;		//data number
	ptr++;

	number = addr + number;

	for(;addr < number;addr++)		//start y,x
	{
		Data = 0;
		Data=*ptr;
		Data<<=8;
		Data|= *(ptr+1);

		AP5200_ImgRegWrite(pDevice,addr,Data);
			  ptr++;
		ptr++;
	}
}

void AP5200_TableLoadSingle( PETHJPG2K pDevice,unsigned char bSlave, unsigned char *ptr)
{
	unsigned char addr, number;
	unsigned short Data;

	addr = *ptr;		//start address
	ptr++;
	number = *ptr;		//data number
	ptr++;

	number = addr + number;

	for(;addr < number;addr++)		//start y,x
	{
		Data = 0;
		Data=*ptr;
		Data<<=8;
		Data|= *(ptr+1);

		AP5200_ImgRegWriteSingle(pDevice,bSlave,addr,Data);
			  ptr++;
		ptr++;
	}
}


///////////////////////////////////////////////////////////////////////////////
//AP5200 RAM font loading  
void AP5200_FontLoad( PETHJPG2K pDevice )
{
	unsigned char i, j, bSlave = 0;
	unsigned char 	*pTable = tFONT_TABLE;
	unsigned short charno = (sizeof(tFONT_TABLE)/sizeof(FONT)); 
	
	//load font enable
     AP5200_ImgRegWrite( pDevice,mon_osd_ctrl,0x0080);
     
     //stop previous cycle if any is active
     WD(pDevice->PLX9056_BAR2[bSlave]+STOP,0);     

	for ( j = 0; j < charno; j++ )
	{
	//	for ( bSlave = 0; bSlave < pDevice->Bridges ; bSlave++)
		{
			//stop previous cycle if any is active
     		WD( pDevice->PLX9056_BAR2[bSlave]+STOP,0);
     
     		//Cycle: W(Addr)-W(0xFA)-W(Row[0])-...-W(Row[7])
     
     		WD( pDevice->PLX9056_BAR2[bSlave]+COMMAND, COMMAND_FONT_WRITE );
     		WD( pDevice->PLX9056_BAR2[bSlave]+DATA, 0xFA );
     		WD( pDevice->PLX9056_BAR2[bSlave]+DATA, j );
     
     		for ( i = 0; i < 8; i++)
     			WD(pDevice->PLX9056_BAR2[bSlave]+DATA,*pTable++);
     
     		WD( pDevice->PLX9056_BAR2[bSlave]+STOP,0);		
		}
	}
	
     //load font disable
	AP5200_ImgRegWrite( pDevice,mon_osd_ctrl,0x0030);  
}

#define MANUALREF 1

unsigned char AP5200_Motion( PETHJPG2K pDevice, AP5200_MOTION_ENABLE_CH *e)
{
  unsigned short Data,Data1,i;

  if ( e->ch > (4*pDevice->Bridges-1) )
    return 0;
  
  /*Select master chip common interface*/
  AP5200_MotRegWrite(pDevice,MDCH_sel,0x0000);
  wait_sync();

  Data = AP5200_MotRegReadSha(pDevice, MD_off);

  /*Turn off motion detection on specified channel*/
  Data |= 0x01 << e->ch;

  AP5200_MotRegWrite(pDevice,MD_off,Data); 

  wait_sync();

  /*If we are disabling interface we have finished*/
  if ( !e->enable )
    {
      /*Turn off the grid*/
      AP5200_MotRegWrite(pDevice,bypass, Data);
      return 1;
    }
  
   
  /*Turn on the grid*/  
  Data = AP5200_MotRegReadSha(pDevice,bypass);
  Data1 = Data | (0x0001<<e->ch);
  Data &= ~(0x0001<<e->ch);
  if (e->show)
    AP5200_MotRegWrite(pDevice,bypass,Data);
  else
    AP5200_MotRegWrite(pDevice,bypass,Data1);
    
  AP5200_MotRegWrite(pDevice,cursor_on,0x0000);	// cursor_off	
  AP5200_MotRegWrite(pDevice,auto_REF,Data1);			// auto refresh on
  if (e->accumulate)
    AP5200_MotRegWrite(pDevice,ACC_sta,Data1);			// accumulate on
  else
    AP5200_MotRegWrite(pDevice,ACC_sta,Data);			// accumulate off

  /*Window colors*/
  AP5200_MotRegWrite(pDevice,selwin,0x3fc0);	// sel_win CB,CR
  AP5200_MotRegWrite(pDevice,mdwin,0xc07f);	// MD_win CB,CR
  AP5200_MotRegWrite(pDevice,cursor,0x1040);	// cursor CB,CR

  /*windows, cursor & interrupt: low level when motion or ABCD */
  AP5200_MotRegWrite(pDevice,mixed_6,0x4E12);	
  

  /*enable interrupt for the channel*/
  Data = AP5200_MotRegReadSha(pDevice,INT_mask);
  Data &= ~(0x0001<<e->ch);
  AP5200_MotRegWrite(pDevice,INT_mask, Data);

  /*select alogics motion filter*/
  Data = AP5200_MotRegReadSha(pDevice,0xfc);
  AP5200_MotRegWrite(pDevice,common_sel, Data & ~(1<<13));

  /*select channel interface*/
  AP5200_MotRegWrite(pDevice,MDCH_sel,(0x0001<<e->ch)); 

  wait_sync();
    
  /*set windows to detect*/
  for ( i = 0; i < 12; i++ ) {
    AP5200_MotRegWrite(pDevice,win_row1_word+i,e->Windows[i]);
  }

  /*select absolute motion threshold*/
  AP5200_MotRegWrite(pDevice,sel_TH_motion_reg,e->sel_TH_motion);			
	
  /*TO DO set the following wisely...*/
  /*Thresholds settings*/
  AP5200_MotRegWrite(pDevice,TH_motion_reg,e->TH_motion);	//Absolute motion detection thresold
  AP5200_MotRegWrite(pDevice,TH_blind_reg,e->TH_blind);	//Blind camera detection threshold
  //AP5200_MotRegWrite(pDevice,mixed_5,0x2020);//No. of detected window to generate an interrupt; 
  AP5200_MotRegWrite(pDevice,TH_alogics,(e->TH_L << 8) | e->TH_H);//Alogics motion filter thresolds
  //#ifdef MANUALREF
  AP5200_MotRegWrite(pDevice,mixed_5,(e->TH_win <<8) | e->REF_catch);//No. of detected window to generate an interrupt; 
  //refresh interval in frame unit for reference image
#if 0
  //TEST for CPLD
  for(i=0; i<5; i++) {
    printk("DELME: a %04x\n", AP5200_MotRegRead(pDevice,TH_alogics, 0));
  }
  for(i=0; i<5; i++) {
    printk("DELME: m %04x\n", AP5200_MotRegRead(pDevice,mixed_5, 0));
  }
  for(i=0; i<5; i++) {
    printk("DELME: a %04x\n", AP5200_MotRegRead(pDevice,TH_alogics, 0));
    printk("DELME: m %04x\n", AP5200_MotRegRead(pDevice,mixed_5, 0));
  }
#endif
  /*Select common register interface*/
  AP5200_MotRegWrite(pDevice,MDCH_sel,0x0000);
  wait_sync();
	
  /*Turn on motion detection on specified channel*/
  Data = AP5200_MotRegReadSha(pDevice, MD_off);
  Data &= ~(0x01<<e->ch);
  AP5200_MotRegWrite(pDevice,MD_off,Data);

  return 1;    
}

// res bit 0-7 => motion status ch 0-7, bit 16-23 => blind status ch 0-7
unsigned char AP5200_GetMotionStatus(PETHJPG2K pDevice, AP5200_MOTION_STATUS *s, int detailed) {
  int i, j;
  unsigned int ChannelOn, Data, Interrupt=0, BlindChannel=0;

  memset(s, 0, sizeof(*s));  

  ChannelOn = AP5200_MotRegReadSha(pDevice,MD_off);

#if 0
  // NOTE: every function must leave MDCH_sel to 0!
  /*Disable motion and interrupts on ALL channels*/
  AP5200_MotRegWrite(pDevice,MDCH_sel,0);
  wait_sync();
#endif

  /*Read interrupt status*/
  if ( pDevice->Bridges == 2 )
    {
      /*select slave chip common interface*/
      Data =  AP5200_MotRegReadSha(pDevice,common_sel);
      AP5200_MotRegWrite(pDevice,common_sel,(Data & 0x3FFF) | (1<<14));
      Interrupt = (UCHAR)AP5200_MotRegRead1(pDevice,INT_status,1) & 0xf;
      BlindChannel = (UCHAR)AP5200_MotRegRead1(pDevice,blind,1) & 0xf;
      Interrupt <<= 4;
      BlindChannel <<=4;
    }
		
  /*select master chip common interface*/
  Data =  AP5200_MotRegReadSha(pDevice,common_sel);
  AP5200_MotRegWrite(pDevice,common_sel,Data & 0x3FFF);
  Interrupt |= ((UCHAR)AP5200_MotRegRead1(pDevice,INT_status,0) & 0xf);
  BlindChannel |= ((UCHAR)AP5200_MotRegRead1(pDevice,blind,0) & 0xf);  
  if (detailed) {
    AP5200_MotRegWrite(pDevice,MD_off,0xFFFF);
    AP5200_MotRegWrite(pDevice,INT_mask,0xFFFF);
    wait_sync();
  }

  /*Save status*/
  for ( i = 0; i < pDevice->Bridges*4; i++ )
    {
      if (!(ChannelOn & (1<<i))) {
	s->ch[i].on = 1;
	if ( Interrupt & (0x01<<i) )
	  s->ch[i].MotionDetected = 1;
	/*Select channel*/
	if ( BlindChannel & (0x01<<i) )
	  s->ch[i].BlindDetected = 1;
	if (detailed) {
	  AP5200_MotRegWrite(pDevice,MDCH_sel,(1<<i));
	  wait_sync();
	  for ( j = 0; j < MAX_GRID_COL; j++ )
	    s->ch[i].Windows[j] = AP5200_MotRegRead(pDevice, MD_status1_word+j, (i>3)?1:0);
	  s->ch[i].sigma = (AP5200_MotRegRead(pDevice,0xd4, (i>3)?1:0)<<8) +
	    (AP5200_MotRegRead(pDevice,0xd5, (i>3)?1:0)>>8);
	}
      }
    }	    
  
  if (detailed) {
    /*Re-enable AP5200 interrupts and motion*/
    AP5200_MotRegWrite(pDevice,MDCH_sel,0);
    wait_sync();
    AP5200_MotRegWrite(pDevice,INT_mask,ChannelOn);
    AP5200_MotRegWrite(pDevice,MD_off,ChannelOn);	
  }

#ifdef MANUALREF
  AP5200_MotRegWrite(pDevice,REF_shot,0xffff);	
#endif
  
  return 1; 
}


unsigned char AP5200_Playback( PETHJPG2K pDevice, AP5200_PLAYBACK *e)
{
  if (! e->enable ) {
    AP5200_ImgRegWrite( pDevice, 0x80, 0 );
  }
  else {
    AP5200_ImgRegWrite( pDevice, 0x80, (1<<2) | 1 ); // force our playback ch is 0 for future use
    AP5200_ImgRegWrite( pDevice, 0x81, 1 ); // playback just ch 0
    AP5200_ImgRegWrite( pDevice, 0x82, 1 ); // connect playbck of ch 0
    AP5200_ImgRegWrite( pDevice, 0x83, e->dest_ch ); // playback of 0 to the given path
    AP5200_ImgRegWrite( pDevice, 0x84, 0 );
    AP5200_ImgRegWrite( pDevice, 0x85, 0 );
    AP5200_ImgRegWrite( pDevice, 0x86, 0 );
    if (e->isNtsc) {
      AP5200_ImgRegWrite( pDevice, 0x87, 0x83b4 );
      AP5200_ImgRegWrite( pDevice, 0x88, 0 );
    }
    else {
      AP5200_ImgRegWrite( pDevice, 0x87, 0x90b4 );
      AP5200_ImgRegWrite( pDevice, 0x88, 0 );
    }
  }
  return 0;
}
