/*
  Copyright 2007 Eurotech S.P.A. (Luca Calligaris)
  Copyright 2007 Christian Pellegrin <chripell@gmail.com> EVOL S.R.L.
  
  This software is released under the GPL v2. Please see the file
  COPYING in the top-level directory.
*/
#ifndef	_AP5200_H
#define _AP5200_H
#include "AP5200_ioctl.h"
#include "driver.h"
/*AP5200 registers offset*/

/*0x00 undefined*/
#define   mixed_1             0x01      /*Sync polarity / Multi-chip settings / ...*/
/*0x02: undefined*/            
#define   mixed_2             0x03      /*MPOUT settings / PictureInPicture */

/*Monitor path*/
#define   mon_ch_on           0x04      /*Activate/deactivate channels*/
#define   mon_ch_loss         0x05      
#define   mon_ch_freeze       0x06      /*Freeze channels*/
/*0x07  undefined*/
#define   MST_mon_sel         0x08      /*Associate input channel to output channel*/
#define   SLV1_mon_sel        0x09
/*0x0A~0x0B: Slave chips*/
#define   MST_ch_scl_mode     0x0C      /*Scale factor for channel displaying on monitor output ( Full, quarter... )*/
#define   SLV1_ch_scl_mode    0x0D
/*0x0E~0x0F: Slave chips*/
#define   ch1_start           0x10      /*[Top:Left] corner of channel displaying on monitor output*/
#define   ch2_start           0x11      /**/
#define   ch3_start           0x12      /**/
#define   ch4_start           0x13      /**/
#define   ch5_start           0x14      /**/
#define   ch6_start           0x15      /**/
#define   ch7_start           0x16      /**/
#define   ch8_start           0x17      /**/
/*0x18~0x1F: Slave chips*/
#define   ch1_length          0x20      /*[Height:Width] of channel displaying on monitor output*/
#define   ch2_length          0x21      /**/
#define   ch3_length          0x22      /**/ 
#define   ch4_length          0x23      /**/
#define   ch5_length          0x24      /**/
#define   ch6_length          0x25      /**/
#define   ch7_length          0x26      /**/ 
#define   ch8_length          0x27      /**/
/*0x28~0x2F: Slave chips*/
#define   ch1_pt              0x30      /*[Vertical:Horizontal] panning of channel displaying on monitor output*/ 
#define   ch2_pt              0x31      /**/ 
#define   ch3_pt              0x32      /**/
#define   ch4_pt              0x33      /**/
#define   ch5_pt              0x34      /**/
#define   ch6_pt              0x35      /**/ 
#define   ch7_pt              0x36      /**/
#define   ch8_pt              0x37      /**/
/*0x38~0x0x3F: Slave chips*/
#define   mon_window_fill     0x40      /*Fill in specified channel area with background color*/      
#define   ch_border_on        0x41      /*Border on/off for specified channel*/
#define   mon_rd_h_start      0x42      /*Adjusting monitor output start position in horizontal direction - MSB must be 0*/
#define   mon_rd_v_start      0x43      /*Adjusting even/odd field monitor output start position in vertical direction*/
#define   mon_rd_valid        0x44      /*Adjusting line width (in vertical direction) and pixel width (in horizontal direction) for monitor output*/
#define   mixed_3             0x45      /*Border settings / Freeze settings / ...*/
#define   mixed_4             0x46      /*Clear SDRAM / Filling options / ...*/
#define   line_back_y         0x47      /*[Border:Background] Y color component*/      
#define   back_cr_cb          0x48      /*[Cr:Cb] color components for background*/      
#define   line_cr_cb          0x49      /*[Cr:Cb] color components for border*/
/*0x4A~0x4F undefined*/

/*Record path*/
#define   rec_main			  0x50
#define   rec_status          0x51
#define   rec_priority        0x52
#define   rec_ch_on           0x53      /*Activate/deactivate channels*/
#define   rec_ch_loss         0x54      
#define   rec_ch_freeze       0x55      /*Freeze channels*/
#define	  rec_length	      0x56
#define	  rec_pt	      0x57
#define   rec_rd_h_start      0x58      /*Adjusting record output start position in horizontal direction - MSB must be 0*/
#define   rec_rd_v_start      0x59      /*Adjusting even/odd field record output start position in vertical direction*/
#define   rec_rd_valid        0x5A      /*Adjusting line width (in vertical direction) and pixel width (in horizontal direction) for record output*/
#define   self_trig	      0x5B
#define	  last_table	      0x5C
#define   rec_sel             0x5E	/*Associate input channel to record channel*/
#define	  rec_table	      0x60	/*record tables 4 nibbles x 8 registers 0x60~0x67*/
/*0x68~0x6F undefined*/
#define	  rec_ch_com	      0x70
#define	  curr_month_year     0x71
#define	  curr_hour_day       0x72
#define	  curr_sec_min        0x73

#define   mon_osd_ctrl        0xA0      /*monitor OSD*/

#define   mon_osd_addr        0xAE
#define   mon_osd_data        0xAF

#define   rec_osd_ctrl        0xB0      /*record OSD*/
#define   MST_rec_osd_addr		0xB5
#define   MST_rec_osd_data		0xB6
#define   SLV1_rec_osd_addr		0xB7
#define   SLV1_rec_osd_data		0xB8


#define   zoom_ctrl           0xC0
#define   zoom_start          0xC1
#define   zoom_length         0xC2
#define   view_start          0xC3
#define   view_length         0xC4

/*channel motion interface*/
#define		TH_alogics			0xD0
#define		TH_motion_reg			0xD1
#define		TH_blind_reg			0xD2
#define		mixed_5				0xD3
#define		sigma_h				0xD4
#define		sigma_l				0xD5
#define		win_row1_word       0xD7
/*...*/
#define		win_row12_word		0xE2
#define		MD_status1_word		0xE3
/*...*/
#define		MD_status12_word	0xEE
#define		sel_TH_motion_reg		0xEF

/*common motion interface*/
#define		MD_off				0xF0 /*channel motion enable*/
#define		bypass				0xF1
#define		cursor_on			0xF2
#define		auto_REF			0xF3
#define		REF_shot			0xF4
#define		ACC_sta				0xF5
#define		selwin				0xF6
#define		mdwin				0xF7
#define		cursor				0xF8
#define		mixed_6				0xF9
#define		MDCH_sel			0xFA /*common/channel register selection*/
#define		INT_mask			0xFB
#define		common_sel			0xFC /*half reg is write only...*/
#define		INT_status			0xFC /*half reg is read only*/		
#define		blind				0xFD				


/*Border thickness allowed values*/
#define	  NO_BORDER		0
#define   NORMAL_BORDER	2
#define   DOUBLE_BORDER	2

/*Zoom coefficients*/
#define	ZOOM_2_0X	0x80
#define ZOOM_2_5X	0x66
#define ZOOM_3_0X	0x55
#define	ZOOM_3_5X	0x49
#define ZOOM_4_0X	0x40

/*Scaling factors  SCALE_HORIZONTAL_X_VERTICAL*/				
#define	SCALE_1_X_1			0	/*Full*/
#define	SCALE_1_2_X_1_2		1	/*Quarter*/
#define	SCALE_1_3_X_1_3		2	/*Third*/
#define	SCALE_1_4_X_1_4		3	/*Sixteenth*/
#define	SCALE_2_3_X_2_3		4	/*Channel 1,5,9,13 only*/
#define	SCALE_1_2_X_1		5	
#define	SCALE_1_4_X_1_2		6

#define X_DISPLAY_UNIT      4
#define	Y_DISPLAY_UNIT      2

#define   MONITOR_PATH		0
#define   RECORD_MST_PATH   (MONITOR_PATH+1)
#define   RECORD_SLV_PATH   (MONITOR_PATH+2)


#define OSD_COLS        34  
#define OSD_ROWS        22        
#define OSD_COLOR_DEF  	0x000
#define OSD_COLOR_A     0x100
#define OSD_COLOR_B     0x200
#define OSD_COLOR_C    	0x300

#define	PAL		0
#define	NTSC	1


#define PAL_FULL_DIM	0x90B2	/*(y,x)*/
#define NTSC_FULL_DIM	0x78B2

#define PAL_QUAD_DIM	0x4859
#define NTSC_QUAD_DIM	0x3C59

#define PAL_ALL_DIM		0x303B
#define NTSC_ALL_DIM	0x283B	

#define	PAL_WIDTH	0xB2
#define	PAL_HEIGHT	0x90

#define	NTSC_WIDTH	0xB2
#define	NTSC_HEIGHT	0x78



/*High level functions*/
unsigned char	AP5200_Init( PETHJPG2K pDevice, unsigned char bNtsc );
unsigned char 	AP5200_SetMode( PETHJPG2K pDevice, PAP5200_MODE_PARAMS pParams);
unsigned char   AP5200_Motion( PETHJPG2K pDevice, AP5200_MOTION_ENABLE_CH *e);
unsigned char   AP5200_GetMotionStatus(PETHJPG2K pDevice, AP5200_MOTION_STATUS *s, int detailed);
unsigned char	AP5200_Zoom ( PETHJPG2K pDevice, PAP5200_RECT Zoom , PAP5200_RECT View, 
								unsigned char ZoomCoeff, unsigned char bEnableZoom );
void			AP5200_OSDEnable( PETHJPG2K pDevice, unsigned char TargetPath, unsigned char bEnable, 
								AP5200_COLOR OsdColorA, AP5200_COLOR OsdColorB, AP5200_COLOR OsdColorC);
unsigned char	AP5200_OSDWriteString( PETHJPG2K pDevice, unsigned char TargetPath,
								unsigned char x, unsigned char y, char *String, unsigned short Color );	  			
void			AP5200_OSDClear( PETHJPG2K pDevice, unsigned char TargetPath );

/*Low level functions*/
void AP5200_ImgRegWriteSingle( PETHJPG2K pDevice, unsigned char bSlave, unsigned char Address, unsigned short Data );
void AP5200_ImgRegWrite( PETHJPG2K pDevice, unsigned char Address, unsigned short Data );
unsigned short	AP5200_ImgRegReadSha( PETHJPG2K pDevice, unsigned char Address );
unsigned short	AP5200_MotRegRead( PETHJPG2K pDevice, unsigned char Address, unsigned char bSlave);
unsigned short	AP5200_MotRegReadSha( PETHJPG2K pDevice, unsigned char Address );
void AP5200_MotRegWrite( PETHJPG2K pDevice, unsigned char Address, unsigned short Data );
void AP5200_TableLoadSingle( PETHJPG2K pDevice,unsigned char bSlave, unsigned char *ptr);
void AP5200_TableLoad( PETHJPG2K pDevice, unsigned char *pTable);
void AP5200_FontLoad( PETHJPG2K pDevice );
void AP5200_EraseSdram(PETHJPG2K pDevice );
unsigned char AP5200_Playback( PETHJPG2K pDevice, AP5200_PLAYBACK *e);

#endif //_AP5200_H

