/*
  Copyright 2007 Eurotech S.P.A. (Luca Calligaris)
  Copyright 2007 Christian Pellegrin <chripell@gmail.com> EVOL S.R.L.
  
  This software is released under the GPL v2. Please see the file
  COPYING in the top-level directory.
*******************************************************************************/

#include "I2C.h"
#include "driver.h"
#include "utils.h"

#include "I2C_T.c"	//init tables

//Internal functions

//SAA7113H decoders mapping {Mux channel, I2C address}
unsigned char	SAA7113H[][2] = {	{4,0x48}, {4,0x4A},{5,0x48}, {5,0x4A},
									{6,0x48}, {6,0x4A},{7,0x48}, {7,0x4A} };

void PCA9564_WriteReg(PETHJPG2K pDevice, unsigned char Register, unsigned char Data)
{
	WD( pDevice->PLX9056_BAR2[0]+Register, Data );
}
unsigned char  PCA9564_ReadReg( PETHJPG2K pDevice, unsigned char Register) 
{
	return ( (unsigned char)( RD(pDevice->PLX9056_BAR2[0]+Register) & 0xFFL ) );
}

unsigned char PCA9564_CheckStatus(PETHJPG2K pDevice, unsigned char StatusCode)
{
	unsigned short  Timeout = I2C_TIMEOUT;

	//Poll interrupt flag
	do
	{
		if ( PCA9564_ReadReg(pDevice,I2CCON) & SI )
			break;
	}
	while (--Timeout);

	if ( !Timeout )
		return 0;

	Timeout = I2C_TIMEOUT;

	do
	{
		if (  PCA9564_ReadReg(pDevice,I2CSTA) == StatusCode )
			break;
		Sleep(1);
	}while( --Timeout );

#ifdef DEBUG_COMMON
	if (!Timeout) {
	  DEBUGP("%s: timeount on status code %d\n", __FUNCTION__, StatusCode);
	}
#endif

	return (unsigned char)Timeout;
}

unsigned char  PCA9564_Init(PETHJPG2K pDevice)
{
	unsigned short  Timeout = I2C_TIMEOUT;
	
	PCA9564_WriteReg(pDevice,I2CCON, I2C_CMD_MASK);

	do
	{
		if (  PCA9564_ReadReg(pDevice,I2CSTA) == NOP )
			break;
		Sleep(1);
	}while( --Timeout );

#ifdef DEBUG_COMMON
	if (!Timeout) {
	  DEBUGP("%s: timeount on init\n", __FUNCTION__);
	}
#endif

	return (unsigned char)Timeout;
}

unsigned char I2C_TxByte(PETHJPG2K pDevice,unsigned char I2CID, unsigned char Data)
{
	unsigned short  Timeout = I2C_TIMEOUT;

	//Start
	PCA9564_WriteReg(pDevice,I2CCON, I2C_CMD_MASK | STA);

	if ( !PCA9564_CheckStatus(pDevice,STA_TX) )
	{

		return 0;
	}

	//Device ID
	PCA9564_WriteReg(pDevice,I2CDAT, I2CID);

	//Clear SI
	PCA9564_WriteReg(pDevice,I2CCON, I2C_CMD_MASK);

	if ( !PCA9564_CheckStatus(pDevice,I2C_WR_ACK) )
	{

		return 0;
	}

	//Immediate Data
	PCA9564_WriteReg(pDevice,I2CDAT, Data);

	//Clear SI
	PCA9564_WriteReg(pDevice,I2CCON, I2C_CMD_MASK);

	if ( !PCA9564_CheckStatus(pDevice,DATA_TX_ACK) )
	{

		return 0;
	}

	//Stop
	PCA9564_WriteReg(pDevice,I2CCON, I2C_CMD_MASK | STO);

	do
	{
		if (  PCA9564_ReadReg(pDevice,I2CSTA) == NOP )
			break;
		Sleep(1);
	}while( --Timeout );

#ifdef DEBUG_COMMON
	if (!Timeout) {
	  DEBUGP("%s: timeount on stop\n", __FUNCTION__);
	}
#endif

	return (unsigned char)Timeout;
}

unsigned char I2C_WriteByte(PETHJPG2K pDevice, unsigned char I2CID, unsigned char SubAddress, unsigned char Data)
{
	unsigned short  Timeout = I2C_TIMEOUT;
	
	//Start
	PCA9564_WriteReg(pDevice,I2CCON, I2C_CMD_MASK | STA);

	if ( !PCA9564_CheckStatus(pDevice,STA_TX) )
	{

		return 0;
	}


	//Device ID
	PCA9564_WriteReg(pDevice,I2CDAT, I2CID);

	//Clear SI
	PCA9564_WriteReg(pDevice,I2CCON, I2C_CMD_MASK);

	if ( !PCA9564_CheckStatus(pDevice,I2C_WR_ACK) )
	{
		return 0;
	}


	//Device Subaddress
	PCA9564_WriteReg(pDevice,I2CDAT, SubAddress);

	//Clear SI
	PCA9564_WriteReg(pDevice,I2CCON, I2C_CMD_MASK);

	if ( !PCA9564_CheckStatus(pDevice,DATA_TX_ACK) )
	{
		return 0;
	}

	//Data
	PCA9564_WriteReg(pDevice,I2CDAT, Data);

	//Clear SI
	PCA9564_WriteReg(pDevice,I2CCON, I2C_CMD_MASK);

	if ( !PCA9564_CheckStatus(pDevice,DATA_TX_ACK) )
	{
		return 0;
	}

	//Stop
	PCA9564_WriteReg(pDevice,I2CCON, I2C_CMD_MASK | STO);

	do
	{
		if (  PCA9564_ReadReg(pDevice,I2CSTA) == NOP )
			break;
		Sleep(1);
	}while( --Timeout );

#ifdef DEBUG_COMMON
	if (!Timeout) {
	  DEBUGP("%s: timeount on stop\n", __FUNCTION__);
	}
#endif

	return (unsigned char)Timeout;
}

unsigned char  I2C_ReadByte(PETHJPG2K pDevice, unsigned char I2CID, unsigned char SubAddress, unsigned char* pData)
{
	unsigned short  Timeout = I2C_TIMEOUT;
	
	//Start
	PCA9564_WriteReg(pDevice,I2CCON, I2C_CMD_MASK | STA);

	if ( !PCA9564_CheckStatus(pDevice,STA_TX) )
		return 0;

	//Device ID
	PCA9564_WriteReg(pDevice,I2CDAT, I2CID);

	//Clear SI
	PCA9564_WriteReg(pDevice,I2CCON, I2C_CMD_MASK);

	if ( !PCA9564_CheckStatus(pDevice,I2C_WR_ACK) )
		return 0;

	//Device Subaddress
	PCA9564_WriteReg(pDevice,I2CDAT, SubAddress);

	//Clear SI
	PCA9564_WriteReg(pDevice,I2CCON, I2C_CMD_MASK);

	if ( !PCA9564_CheckStatus(pDevice,DATA_TX_ACK) )
		return 0;

	//Clear SI & repeated start
	PCA9564_WriteReg(pDevice,I2CCON, I2C_CMD_MASK | STA);

	if ( !PCA9564_CheckStatus(pDevice,REP_STA) )
		return 0;

	//Device ID
	PCA9564_WriteReg(pDevice,I2CDAT, I2CID | I2C_READ);

	//Clear SI
	PCA9564_WriteReg(pDevice,I2CCON, I2C_CMD_MASK);

	if ( !PCA9564_CheckStatus(pDevice,I2C_RD_ACK) )
		return 0;

	//Clear SI, wait for data then send ACK
	PCA9564_WriteReg(pDevice,I2CCON, I2C_CMD_MASK | AA);

	//Wait for status ok
	if ( !PCA9564_CheckStatus(pDevice,DATA_RX_ACK) )
		return 0;


	//Read data
	*pData = PCA9564_ReadReg(pDevice,I2CDAT);

	//Clear SI and Stop
	PCA9564_WriteReg(pDevice,I2CCON, I2C_CMD_MASK | STO);

	do
	{
		if (  PCA9564_ReadReg(pDevice,I2CSTA) == NOP )
			break;
		Sleep(1);
	}while( --Timeout );

#ifdef DEBUG_COMMON
	if (!Timeout) {
	  DEBUGP("%s: timeount on stop\n", __FUNCTION__);
	}
#endif

	return (unsigned char)Timeout;
}



//Exported functions

unsigned char  SAA7113H_Init(PETHJPG2K pDevice, unsigned char Decoder , unsigned char bNtsc)
{
	unsigned char *pTable = bNtsc ? tDECODER_NTSC : tDECODER_PAL;
	unsigned char address = *pTable++;
	unsigned char loop = *pTable++;
	unsigned short i;

	//Select mux channel
	if ( !I2C_TxByte(pDevice, I2CID_MUX, SAA7113H[Decoder][0]))
		return 0;

	for(i = 0 ; i < loop; i++)
	{
		if ( !I2C_WriteByte(pDevice,SAA7113H[Decoder][1], address+i, *pTable++))
			return 0;
	}

	return 1;
}

unsigned char  SAA7113H_WriteByte(PETHJPG2K pDevice, unsigned char Decoder, unsigned char SubAddress, unsigned char Data)
{
	//Select mux channel
	if ( !I2C_TxByte(pDevice, I2CID_MUX, SAA7113H[Decoder][0]))
		return 0;
	//Write data
	
	return I2C_WriteByte(pDevice,SAA7113H[Decoder][1], SubAddress, Data);
}

unsigned char SAA7113H_ReadByte(PETHJPG2K pDevice, unsigned char Decoder, unsigned char SubAddress, unsigned char* pData)
{
	//Select mux channel
	if ( !I2C_TxByte(pDevice, I2CID_MUX, SAA7113H[Decoder][0]))
		return 0;
	//Read data
	return I2C_ReadByte(pDevice,SAA7113H[Decoder][1], SubAddress, pData);
}



unsigned char SAA7121H_Init(PETHJPG2K pDevice, unsigned char bNtsc)
{
	unsigned char *pTable = bNtsc ? tSAA7121_NTSC1 : tSAA7121_PAL1;
	unsigned char address = *pTable++;
	unsigned char loop = *pTable++;
	unsigned char i;


	for(i = 0 ; i < loop; i++)
	{
		
		if ( !I2C_WriteByte(pDevice,I2CID_SAA7121H,address+i, *pTable++))
			return 0;
	}

	pTable = bNtsc ? tSAA7121_NTSC2 : tSAA7121_PAL2;

	address = *pTable++;
	loop = *pTable++;

	for(i = 0 ; i < loop; i++)
	{
		if ( !I2C_WriteByte(pDevice,I2CID_SAA7121H,address+i, *pTable++))
			return 0;
	}

	//Color bar off
	return I2C_WriteByte(pDevice,I2CID_SAA7121H,0x3A,0x13);

}







