/*
  Copyright 2007 Eurotech S.P.A. (Luca Calligaris)
  Copyright 2007 Christian Pellegrin <chripell@gmail.com> EVOL S.R.L.
  
  This software is released under the GPL v2. Please see the file
  COPYING in the top-level directory.
*/

//#define DEBUG 1

//#define DEBUG_IO 1

#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
#include <linux/config.h>
#endif
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/timer.h>
#include <linux/list.h>
#include <linux/interrupt.h>  
#include <linux/usb.h>
#include <linux/pci.h>
#include <linux/poll.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/proc_fs.h>
#include <linux/fcntl.h>
#include <linux/seq_file.h>
#include <linux/cdev.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/system.h>
#include <asm/unaligned.h>
#include <asm/dma.h>

#include "plx9056.h"
#include "driver.h"

#include "ADV202.H"
#include "AP5200.h"
#include "I2C.h"
#include "plx9056_registers.h"


extern int vbuffer_pages;
extern int vbuffer_images;

/* miscellanous buffers for ioctl */
static AP5200_MODE_PARAMS ap5200_mode_params;
static unsigned char cbuf[2];
static unsigned long firmware[32768/4];
static ETHJPG2K_FIRMWARE_PARAMS firmware_parameters;
static ETHJPG2K_FIRMWARE_DATA firmware_data;
static AP5200_COLOR Color[3];
static AP5200_OSD_PARAMS OsdParams;
static char * OsdString = NULL;
static AP5200_COLOR ColorNull = {0,0,0};
static AP5200_MOTION_ENABLE_CH ap_motion_enable;
static AP5200_MOTION_STATUS ap_motion_status;
static AP5200_PLAYBACK ap_video_playback;


static int restart_adv202(struct plx_dma * pdev) {
	int ndev;
	int type = 0;
	unsigned short id = 0;
	int force = 0;
	
	if (pdev->plx->adv_started)
		return 0;

	pdev->plx->adv_started = 1;

	switch(pdev->plx->mode) {
	case PLX_ENCODE: type = FW_VIDEO_ENCODE; id = 0xff82; break;
	case PLX_DECODE: type = FW_VIDEO_DECODE; id = 0xffa2; break;
	case PLX_HIPI_ENCODE: type = FW_HIPI_ENCODE; id = 0xff82; break;
	case PLX_HIPI_DECODE: type = FW_HIPI_DECODE; id = 0xffa2; break;
	case PLX_HIPI_LOOP: type = FW_HIPI_LOOP; id = 0; break;
	default:
		printq("bug @ %s %d\n", __FUNCTION__, __LINE__);
	}

	if (pdev->plx->mode == PLX_HIPI_ENCODE || pdev->plx->mode == PLX_HIPI_DECODE)
	  force = 0x0084;

	ndev = driver_get_n(pdev->plx);
	pr_debug("restarting ADV202 n. %d\n", ndev);
	pethjpg2k->FirmwareType[ndev] = type;
	ADV202_Reset(pethjpg2k, ndev);
	ADV202_Init(pethjpg2k, ndev, 0, force);
	if (pdev->plx->mode != PLX_HIPI_LOOP) {
		ADV202_LoadFirmware(pethjpg2k, ndev, firmware, 32768/4);
		ADV202_LoadFirmwareParameters(pethjpg2k, ndev, type, &firmware_parameters);
	}
	ADV202_SetupDMA(pethjpg2k, ndev);

	if (pdev->plx->mode == PLX_HIPI_LOOP) {
		ADV202_WriteIndirect(pethjpg2k, ndev, 0xFFFF1418, 0x00100000);
		if (ADV202_ReadIndirect(pethjpg2k, ndev, 0xFFFF1418) != 0x00100000)
			printq("error on settingup LOOP mode!\n");
	}
	else {
		if (!ADV202_CheckApplicationID(pethjpg2k, ndev, id)) {
			printq("ID mishmash after dnload!\n");
			return -EBADF;
		}
		ADV202_StartExecution(pethjpg2k, ndev);
	}
		
	return 0;
}

static void init_board(struct plx_dma * pdev, int is_ntsc) {
	int ndev;

	pethjpg2k->Flags.bNtsc = is_ntsc;
	ndev = driver_get_n(pdev->plx);
	pr_debug("%s: %d\n", __FUNCTION__, ndev);

	if ( !PCA9564_Init(pethjpg2k)) {
		printq("PCA9564 init failed!\n");
	}

	if ( !SAA7121H_Init(pethjpg2k,is_ntsc) ) {
		printq("SAA7121H init failed!\n");
	}

	if ( !AP5200_Init(pethjpg2k,is_ntsc) ) {
		printq("AP5200 init failed!\n");
	}

}

void smup_setup(struct plx_dma * pdev) {
	struct plx_image *img;
	
	if (pdev->opened & PLX_RUNNING) {
		if (pdev->plx->mode == PLX_HIPI_DECODE || pdev->plx->mode == PLX_HIPI_LOOP) {
			img = &pdev->img[pdev->phead];
			img->hdr.field = 0xffffffff;
			img->hdr.index = 0;
			img->hdr.Info_u.info = 0;
			img->hdr.size = pdev->plx->image_size/4;
			if (alloc_dma_memory_image(img, pdev->plx->image_size) == -1) {
				printq("BUG: buffer ring overflow, increase video_pages\n");
				pdev->overflow++;
				return;
			}
		
			start_dma(pdev, img, 0, 0);
		}
		else {
			smup_start(pdev);
		}
	}
}

static int start_encdec(struct plx_dma * pdev) {
	int i;
	pr_debug("%s %d\n", __FUNCTION__, __LINE__);

	pdev->phead = pdev->ptail = pdev->vhead = pdev->vtail = 0;
	pdev->is_dma = 0;
	pdev->is_dma_hdr = 0;
	for(i=0; i<vbuffer_images; i++)
		pdev->img[i].status = IMAGE_IDLE;

	restart_adv202(pdev);
	pdev->opened |= PLX_RUNNING;

	if ((pdev->opened & PLX_OPENED_IN) == PLX_OPENED_IN) {
		plx_irqon(pdev, PLX_INTCTRLSTS_DMA0_INT_ENA);
		smup_setup(pdev);
	}
	else if ((pdev->opened & PLX_OPENED_OUT) == PLX_OPENED_OUT) {
		plx_irqon(pdev, PLX_INTCTRLSTS_DMA1_INT_ENA);
	}
	else
		printk("BUG %s %d\n", __FUNCTION__, __LINE__);

	return 0;
}

void stop_encdec(struct plx_dma * pdev) {
	pr_debug("%s %d\n", __FUNCTION__, __LINE__);
	pdev->opened &= ~PLX_RUNNING;
	pdev->plx->adv_started = 0;
	if ((pdev->opened & PLX_OPENED_IN) == PLX_OPENED_IN) {
		plx_irqoff(pdev, PLX_INTCTRLSTS_DMA0_INT_ENA);
		udelay(100);
		if (pdev->is_dma 
		    || pdev->is_dma_hdr
		    ) {
			stop_dma(pdev, 0);
			if (pdev->is_dma) {
				handle_data_encode(pdev, 1);
			}
			udelay(100);
		}
	}
	else if ((pdev->opened & PLX_OPENED_OUT) == PLX_OPENED_OUT) {
		plx_irqoff(pdev, PLX_INTCTRLSTS_DMA1_INT_ENA);
		udelay(100);
		if (pdev->is_dma) {
			stop_dma(pdev, 1);
			udelay(100);
			tasklet_enable(&pdev->task);
		}
		while (atomic_read(&pdev->task.count) != 0)
			tasklet_enable(&pdev->task); /* TODO: correct this shit */
		schedule();	/* give tasklet a chance to run */
			
		if (atomic_read(&pdev->task.count) != 0)
			printk("BUG %s %d\n", __FUNCTION__, __LINE__);			
		tasklet_kill(&pdev->task);
	}
	else
		printk("BUG %s %d\n", __FUNCTION__, __LINE__);
}

int plx_ioctl(struct inode *inode, struct file *filp,
                 unsigned int cmd, unsigned long arg)
{
	struct plx_dma *pdev = (struct plx_dma *) filp->private_data;
	//struct plx_dev *bSlave = pdev->plx;
	int ret = -ENOTTY;
	int t;

	switch (cmd) {

		/*VIDEO IN/OUT SETTINGS*/		

	case IOCTL_ETHJPG2K_VIDEO_SET_MODE: 
		if (copy_from_user(&ap5200_mode_params, (PAP5200_MODE_PARAMS)arg, sizeof(AP5200_MODE_PARAMS)))
			return -EFAULT;
		if (AP5200_SetMode(pethjpg2k, &ap5200_mode_params) == 1)
			ret = 0;
		else 
			ret = -EINVAL;
		break;
	case IOCTL_ETHJPG2K_VIDEO_SET_BRIGHTNESS: 
		if (copy_from_user(&cbuf[0], (unsigned char *)arg, 2))
			return -EFAULT;
		SAA7113H_WriteByte(pethjpg2k,cbuf[0],SAA7113H_BRIGHT,cbuf[1]);
		ret = 0;
		break;
	case IOCTL_ETHJPG2K_VIDEO_SET_CONTRAST: 
		if (copy_from_user(&cbuf[0], (unsigned char *)arg, 2))
			return -EFAULT;
		SAA7113H_WriteByte(pethjpg2k,cbuf[0],SAA7113H_CONTRAST,cbuf[1]);
		ret = 0;
		break;
	case IOCTL_ETHJPG2K_VIDEO_SET_HUE: 
		if (copy_from_user(&cbuf[0], (unsigned char *)arg, 2))
			return -EFAULT;
		SAA7113H_WriteByte(pethjpg2k,cbuf[0],SAA7113H_HUE,cbuf[1]);
		ret = 0;
		break;

	case IOCLT_ETHJPG2K_VIDEO_CLEAR:
		AP5200_EraseSdram(pethjpg2k);
		msleep(21);
		ret = 0;
		break;

	case IOCLT_ETHJPG2K_VIDEO_PLAYBACK:
		if (copy_from_user(&ap_video_playback, (AP5200_PLAYBACK *) arg, sizeof(ap_video_playback)))
			return -EFAULT;
		AP5200_Playback(pethjpg2k, &ap_video_playback);
		ret = 0;
		break;


		/*OSD settings */
		/*( dwCode-IOCTL_ETHJPG2K_MONITOR_XXX_OSD )/0x10 == ( MONITOR_PATH, RECORD_MST_PATH, RECORD_SLV_PATH )
			THIS WORKS ONLY BEACUSE OF IOCTL DEFINITIONS!!!*/

	case IOCTL_ETHJPG2K_MONITOR_ENABLE_OSD: 
	case IOCTL_ETHJPG2K_RECORD_MST_ENABLE_OSD: 
	case IOCTL_ETHJPG2K_RECORD_SLV_ENABLE_OSD: 
		if (copy_from_user(&Color[0], (AP5200_COLOR *)arg, sizeof(AP5200_COLOR)*3))
			return -EFAULT;			
		AP5200_OSDEnable(pethjpg2k,(unsigned char)((cmd-IOCTL_ETHJPG2K_MONITOR_ENABLE_OSD)/0x10),TRUE,Color[0],Color[1],Color[2]);				
		ret = 0;
		break;   	

	case IOCTL_ETHJPG2K_MONITOR_DISABLE_OSD:
	case IOCTL_ETHJPG2K_RECORD_MST_DISABLE_OSD:
	case IOCTL_ETHJPG2K_RECORD_SLV_DISABLE_OSD:		
		AP5200_OSDEnable(pethjpg2k,(unsigned char)((cmd-IOCTL_ETHJPG2K_MONITOR_DISABLE_OSD)/0x10),FALSE,ColorNull,ColorNull,ColorNull);
		ret = 0;
		break;   			
		
	case IOCTL_ETHJPG2K_MONITOR_CLEAR_OSD: 
	case IOCTL_ETHJPG2K_RECORD_MST_CLEAR_OSD: 
	case IOCTL_ETHJPG2K_RECORD_SLV_CLEAR_OSD: 
		AP5200_OSDClear( pethjpg2k, (unsigned char)((cmd-IOCTL_ETHJPG2K_MONITOR_CLEAR_OSD)/0x10) );
		ret = 0;
		break;  	
		
	case IOCTL_ETHJPG2K_MONITOR_WRITE_OSD: 
	case IOCTL_ETHJPG2K_RECORD_MST_WRITE_OSD: 
	case IOCTL_ETHJPG2K_RECORD_SLV_WRITE_OSD: 
		if (copy_from_user(&OsdParams, (AP5200_OSD_PARAMS *)arg, sizeof(AP5200_OSD_PARAMS)))
			return -EFAULT;			
		if (!OsdString)
			OsdString = (char *) get_zeroed_page(GFP_KERNEL);
		if (OsdParams.StrLen > PAGE_SIZE-1)
			OsdParams.StrLen = PAGE_SIZE-1;
		if (copy_from_user(OsdString, OsdParams.String, OsdParams.StrLen))
			return -EFAULT;			
		OsdString[OsdParams.StrLen] = '\0';
		OsdParams.String = OsdString;
		
		if (AP5200_OSDWriteString(pethjpg2k, 
					  (unsigned char)((cmd-IOCTL_ETHJPG2K_MONITOR_WRITE_OSD)/0x10), 
					  OsdParams.x, OsdParams.y, OsdString, OsdParams.Color)) {
			ret = 0;
		}
		else
			ret = -EINVAL;
		break; 	

		/* controls slave path: 1 enable recording on slave ADV202 */
	case IOCTL_ETHJPG2K_SET_SLAVE_PATH:
		if (copy_from_user(&t, (unsigned int *)arg, sizeof(t)))
			return -EFAULT;
		if (t) {
			unsigned long	Data = RD( pethjpg2k->PLX9056_BAR2[0]+CPLD_STATUS );
			WD( pethjpg2k->PLX9056_BAR2[0]+CPLD_STATUS, Data & ~2);
		}
		else {
			unsigned long	Data = RD( pethjpg2k->PLX9056_BAR2[0]+CPLD_STATUS );
			WD( pethjpg2k->PLX9056_BAR2[0]+CPLD_STATUS, Data | 2);
		}
		ret = 0;
		break;

		/* e0904 management */

	case IOCTL_ETHJPG2K_LINUX_INIT:
		init_board(pdev, (int) arg);
		ret = 0;
		break;		

	case IOCTL_ETHJPG2K_ENCDEC_LOAD_FIRMWARE:
		if (copy_from_user(&firmware_data, (unsigned char *) arg, sizeof(firmware_data)))
			return -EFAULT;

		switch (firmware_data.Type) {
		case FW_HIPI_ENCODE:
			pdev->plx->mode = PLX_HIPI_ENCODE;
			break;
		case FW_HIPI_DECODE:	
			pdev->plx->mode = PLX_HIPI_DECODE;
			break;
		case FW_HIPI_LOOP:
			pdev->plx->mode = PLX_HIPI_LOOP;
			break;
		}
		
		if (pdev->plx->mode != PLX_HIPI_LOOP) {
			if (copy_from_user(firmware, firmware_data.pFirmware, 32768))
				return -EFAULT;
			if (copy_from_user(&firmware_parameters, firmware_data.pFirmwareParameters , 
					   sizeof(firmware_parameters)))
				return -EFAULT;
			if (pdev->plx->mode == PLX_ENCODE) {
				pdev->skip = firmware_parameters.EncodeParams.STALLPAR;
				pdev->is_ntsc = (firmware_parameters.EncodeParams.VFORMAT == 0 ||
						 firmware_parameters.EncodeParams.VFORMAT == 8) ? 1 : 0;
				pdev->is_interlaced = (firmware_parameters.EncodeParams.VFORMAT == 8 ||
						       firmware_parameters.EncodeParams.VFORMAT == 9) ? 1 : 0;
				pdev->coding_style = firmware_parameters.EncodeParams.COD_STYLE;
			}
			if (pdev->plx->mode == PLX_HIPI_DECODE) {
				pdev->plx->image_size = firmware_parameters.HipiParams.XTOT
					* firmware_parameters.HipiParams.YTOT; 
			}
			if (pdev->plx->mode == PLX_HIPI_ENCODE) {
				pdev->skip = firmware_parameters.HipiParams.sub.EncodeParams.STALLPAR;
				pdev->is_ntsc = (firmware_parameters.HipiParams.sub.EncodeParams.VFORMAT == 0 ||
						 firmware_parameters.HipiParams.sub.EncodeParams.VFORMAT == 8) ? 1 : 0;
				pdev->is_interlaced = (firmware_parameters.HipiParams.sub.EncodeParams.VFORMAT == 8 ||
						       firmware_parameters.HipiParams.sub.EncodeParams.VFORMAT == 9) ? 1 : 0;
				pdev->coding_style = firmware_parameters.HipiParams.sub.EncodeParams.COD_STYLE;
			}			
		}
		else {
			pdev->plx->image_size = firmware_data.FirmwareSizeInDwords; /* note: misleading, no Dwords but bytes */
		}
		ret = 0;
		break;
		
	case IOCTL_ETHJPG2K_ENCDEC_START:
		return start_encdec(pdev);
		break;

	case IOCTL_ETHJPG2K_ENCDEC_STOP:
		stop_encdec(pdev);
		ret = 0;
		break;		

	case IOCTL_ETHJPG2K_LINUX_GETBSIZE: {
		unsigned long size = vbuffer_pages;

		if (copy_to_user((unsigned long *) arg, &size, sizeof(unsigned long))) 
			return -EFAULT;			      		
		ret = 0;
	}
	  break;		

	  /* motion detection */

	case IOCTL_ETHJPG2K_MOTION_ENABLE:
	case IOCTL_ETHJPG2K_MOTION_DISABLE:
	  if (copy_from_user(&ap_motion_enable, (AP5200_MOTION_ENABLE_CH *) arg, sizeof(ap_motion_enable)))
	    return -EFAULT;
	  if (cmd == IOCTL_ETHJPG2K_MOTION_ENABLE)
	    ap_motion_enable.enable = 1;
	  else
	    ap_motion_enable.enable = 0;

	  AP5200_Motion(pethjpg2k, &ap_motion_enable);
	  ret = 0;
	  break;

	case IOCTL_ETHJPG2K_MOTION_GET_STATUS:
	case IOCTL_ETHJPG2K_MOTION_GET_STATUS_FAST:
	  AP5200_GetMotionStatus(pethjpg2k, &ap_motion_status, cmd == IOCTL_ETHJPG2K_MOTION_GET_STATUS);
	  if (copy_to_user((AP5200_MOTION_STATUS *) arg, &ap_motion_status, sizeof(ap_motion_status))) 
	    return -EFAULT;			      			  
	  ret = 0;
	  break;

	  /* GPIOs */

	case IOCTL_ETHJPG2K_GPIO_READ:
	  if (copy_from_user(&cbuf[0], (unsigned char *) arg, 2*sizeof(unsigned char)))
	    return -EFAULT;
	  if (cbuf[0] > 1)
	    return -EINVAL;
	  cbuf[1] = (unsigned char) (RD(pethjpg2k->PLX9056_BAR2[cbuf[0]]+GPIO_IN) & 0xFFL );
	  if (copy_to_user((unsigned char *) arg, &cbuf[0], 2*sizeof(unsigned char))) 
	    return -EFAULT;			      			  
	  ret = 0;
	  break;			
		
	case IOCTL_ETHJPG2K_GPIO_WRITE:			
	  if (copy_from_user(&cbuf[0], (unsigned char *) arg, 2*sizeof(unsigned char)))
	    return -EFAULT;
	  if (cbuf[0] > 1)
	    return -EINVAL;
	  WD(pethjpg2k->PLX9056_BAR2[cbuf[0]]+GPIO_OUT, cbuf[1]);		     
	  ret = 0;
	  break;

	default:
		printq("unknown ioctl %08x\n", cmd);
	}

	return ret;
}
