/*
  Copyright 2007 Eurotech S.P.A. (Luca Calligaris)
  Copyright 2007 Christian Pellegrin <chripell@gmail.com> EVOL S.R.L.
  
  This software is released under the GPL v2. Please see the file
  COPYING in the top-level directory.
*/

//#define DEBUG 1

//#define DEBUG_IO 1

#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
#include <linux/config.h>
#endif
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/timer.h>
#include <linux/list.h>
#include <linux/interrupt.h>  
#include <linux/usb.h>
#include <linux/pci.h>
#include <linux/poll.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/proc_fs.h>
#include <linux/fcntl.h>
#include <linux/seq_file.h>
#include <linux/cdev.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/system.h>
#include <asm/unaligned.h>
#include <asm/dma.h>

#include "plx9056.h"
#include "plx9056_registers.h"

#include "ADV202.H"

extern int vbuffer_images;
extern int vbuffer_pages;

void plx_irqon(struct plx_dma *dma, u32 Mask) {
	unsigned long InterruptReg;
	struct plx_dev *bSlave = dma->plx;

	InterruptReg =  PLX9056_ReadConfigReg( bSlave, PLX_ADDR_INTCTRLSTS, SIZE_32BITS);

	PLX9056_WriteConfigReg( bSlave, PLX_ADDR_INTCTRLSTS, InterruptReg | Mask | (1<<8), SIZE_32BITS);	
}

void plx_irqoff(struct plx_dma *dma, u32 Mask) {
	unsigned long InterruptReg;
	struct plx_dev *bSlave = dma->plx;

	InterruptReg =  PLX9056_ReadConfigReg( bSlave, PLX_ADDR_INTCTRLSTS, SIZE_32BITS);

	PLX9056_WriteConfigReg( bSlave, PLX_ADDR_INTCTRLSTS, InterruptReg & ~Mask, SIZE_32BITS);
}

void handle_hdr(struct plx_dma * dma) {
	struct plx_image *img;

	dma->is_dma_hdr = 0;
	if (dma->opened & PLX_RUNNING) {
		img = &dma->img[dma->phead];
		img->hdr.field = htonl(dma->smup_hdr->Field);
		img->hdr.index = htonl(dma->smup_hdr->Index);
		img->hdr.Info_u.info = htonl(dma->smup_hdr->Info_u.Info);
		img->hdr.size = htonl(dma->smup_hdr->Size);

		pr_debug("queuing %d: %08lx %8ld %08lx %ld\n", dma->phead, img->hdr.field, 
		       img->hdr.index, img->hdr.Info_u.info, img->hdr.size);
		if (img->hdr.size*4 > 1000*1024) {
			printq("BUG: image too big!\n");
			return;
		}					
		if (alloc_dma_memory_image(img, img->hdr.size*4) == -1) {
			printq("BUG: buffer ring overflow, increase video_pages\n");
			dma->overflow++;
			return;
		}
		start_dma(dma, img, 0, 0);
	}
}


void handle_data_encode(struct plx_dma * dma, int force) {
	struct plx_image *img;
	u32 head;

	dma->is_dma = 0;
	head = dma->phead;
	img = &dma->img[head];
	img->status = IMAGE_DMA_DONE;
	if (!force)
		pr_debug("isr: finished slot %d\n", head);
	head++;
	if (head == vbuffer_images)
		head = 0;
	if (head == dma->ptail || force) {
		if (!force) {
			printq("%s: images ring overflow\n", __FUNCTION__);
			dma->overflow++;
		}
		free_dma_memory_head(img);
	}
	else {
		dma->phead = head;
		wake_up_interruptible(&dma->pready);
	}			
	smup_setup(dma);
}

void handle_data_decode(struct plx_dma * dma) {
	struct plx_image *img;

	img = &dma->img[dma->ptail];
	unmap_dma_memory_image(img, img->first, img->n);
	dma->vtail = (dma->vtail + img->n) % vbuffer_pages;
	dma->ptail = (dma->ptail + 1) % vbuffer_images;
	if (dma->phead != dma->ptail)
		tasklet_schedule(&dma->task);
	tasklet_enable(&dma->task);
	dma->is_dma = 0;
	wake_up_interruptible(&dma->pready);
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
irqreturn_t plx_pci_isr(int irq, void *__data, struct pt_regs *r) 
#else
irqreturn_t plx_pci_isr(int irq, void *__data) 
#endif
{
	struct pci_dev *dev = (struct pci_dev *) __data;
	struct plx_dev *bSlave = pci_get_drvdata(dev);
	struct plx_dma * dma;
	unsigned long	PLX9056_IntStatus;
	unsigned char	PLX9056_DmaStatus;

	PLX9056_IntStatus = PLX9056_ReadConfigReg(bSlave, PLX_ADDR_INTCTRLSTS, SIZE_32BITS);

	if ( !(PLX9056_IntStatus & 
	       (PLX_INTCTRLSTS_DMA0_INT_ACTIVE | PLX_INTCTRLSTS_DMA1_INT_ACTIVE | PLX_INTCTRLSTS_LCL_INT_ACTIVE))) {
		return IRQ_NONE;
	}

	pr_debug("%s: IRQ: %08lx\n", __FUNCTION__, PLX9056_IntStatus);

	/*DMA0 interrupt*/
	if ( (PLX9056_IntStatus & PLX_DMA0_INT) == PLX_DMA0_INT )
	{
		// disable IRQ
		PLX9056_WriteConfigReg(bSlave, PLX_ADDR_INTCTRLSTS,
			PLX9056_IntStatus & ~PLX_INTCTRLSTS_DMA0_INT_ENA,
			SIZE_32BITS);

		PLX9056_DmaStatus = (unsigned char)PLX9056_ReadConfigReg(bSlave,PLX_ADDR_DMA0_CMDSTS,SIZE_8BITS);
		PLX9056_WriteConfigReg(bSlave,PLX_ADDR_DMA0_CMDSTS,PLX9056_DmaStatus | PLX_CMDSTS_CLRINT,SIZE_8BITS);

		dma = &bSlave->dma[0];
		if (dma->is_dma_hdr) {
			handle_hdr(dma);
		}
		else {
			handle_data_encode(dma, 0);
		}
		
		// reenable IRQ
		PLX9056_WriteConfigReg(bSlave, PLX_ADDR_INTCTRLSTS, PLX9056_IntStatus, SIZE_32BITS);

	}

	/*DMA1 interrupt*/
	if ( (PLX9056_IntStatus & PLX_DMA1_INT) == PLX_DMA1_INT )
	{

		/*Disable, clear, re-enable*/
		PLX9056_WriteConfigReg(bSlave, PLX_ADDR_INTCTRLSTS,
			PLX9056_IntStatus & ~PLX_INTCTRLSTS_DMA1_INT_ENA,
			SIZE_32BITS);

		PLX9056_DmaStatus = (unsigned char)PLX9056_ReadConfigReg(bSlave,PLX_ADDR_DMA1_CMDSTS,SIZE_8BITS);
		PLX9056_WriteConfigReg(bSlave,PLX_ADDR_DMA1_CMDSTS,PLX9056_DmaStatus | PLX_CMDSTS_CLRINT,SIZE_8BITS);

		dma = &bSlave->dma[1];
		handle_data_decode(dma);

		PLX9056_WriteConfigReg(bSlave, PLX_ADDR_INTCTRLSTS, PLX9056_IntStatus, SIZE_32BITS);
	}

	/*ADV202 or motion detector interrupt */
	if ( (PLX9056_IntStatus & PLX_LOCAL_INT ) == PLX_LOCAL_INT )
	{
		/* CHRI TODO */
		{
			static int said = 0;
			if (!said) {
				printq("WARNING NYI at %s %d\n", __FUNCTION__, __LINE__);
				said = 1;
			}
		}
	}

	return IRQ_HANDLED;
}

int plx_request_irq(struct pci_dev *dev, void *isr_data) 
{
	struct plx_dev *pdev = pci_get_drvdata(dev);
	int result = 0;

	result= request_irq(dev->irq, plx_pci_isr,
			    SA_SHIRQ, pdev->name, isr_data);
	dev_dbg(&dev->dev, "irq %d got %d\n", dev->irq, result);

	return result;
}

void plx_free_irq(struct pci_dev *dev, void *isr_data)
{
	free_irq(dev->irq,isr_data);
}


void decode_tasklet(unsigned long pdev_) {
	struct plx_dma *pdev = (struct plx_dma *) pdev_;
	struct plx_image *img;
	
	if (pdev->opened & PLX_RUNNING) {
		if (pdev->phead == pdev->ptail) {
			printk("BUG @ %s %d\n", __FUNCTION__, __LINE__);
			return;
		}
		img = &pdev->img[pdev->ptail];
		if (pdev->vtail != img->dec.start) {
			printk("BUG @ %s %d: %d != %ld\n", __FUNCTION__, __LINE__,
			       pdev->vtail,  img->dec.start);
		}
		if (pdev->is_dma) {
			printk("BUG @ %s %d\n", __FUNCTION__, __LINE__);
		}
		map_dma_memory_image(img, img->dec.start, img->dec.size);
		tasklet_disable_nosync(&pdev->task);
		start_dma(pdev, img, 1, 0);
	}
}
