/*
  Copyright 2007 Eurotech S.P.A. (Luca Calligaris)
  Copyright 2007 Christian Pellegrin <chripell@gmail.com> EVOL S.R.L.
  
  This software is released under the GPL v2. Please see the file
  COPYING in the top-level directory.
*/

//#define DEBUG 1

//#define DEBUG_IO 1

#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
#include <linux/config.h>
#endif
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/delay.h>
#include <linux/ioport.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/smp_lock.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/timer.h>
#include <linux/list.h>
#include <linux/interrupt.h>  
#include <linux/usb.h>
#include <linux/pci.h>
#include <linux/poll.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/proc_fs.h>
#include <linux/fcntl.h>
#include <linux/seq_file.h>
#include <linux/cdev.h>
#include <linux/workqueue.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <asm/system.h>
#include <asm/unaligned.h>
#include <asm/dma.h>

#include "plx9056.h"
#include "plx9056_registers.h"
#include "driver.h"
#include "ADV202.H"

/* module parameters */

/* for efficency it's better that  vbuffer_pages * PAGE_SIZE >= vbuffer_images max_image_size */
int vbuffer_pages = 1024;
module_param(vbuffer_pages, int, S_IRUGO);
int vbuffer_images = 10;
module_param(vbuffer_images, int, S_IRUGO);

int quiet = 0;
module_param(quiet, int, S_IRUGO);

/* globals */

PETHJPG2K pethjpg2k;

/* kernel version specifics */
#define mem_map_reserve(adr)		SetPageReserved(adr)
#define mem_map_unreserve(adr)		ClearPageReserved(adr)

static inline unsigned long kvirt_to_pa(unsigned long adr)
{
   unsigned long kva;

   kva = (unsigned long) page_address(vmalloc_to_page((void *) adr));
   kva |= adr&(PAGE_SIZE-1);   /* restore the offset */
   return __pa(kva);
}
/* end kernel version specifics */

int plx_major = 0;
module_param(plx_major, int, S_IRUGO);

int plx_minor = 0;
module_param(plx_minor, int, S_IRUGO);


void plx_reg_write32(struct pci_dev *dev,u32 reg,u32 data)
{ 
	struct plx_dev *pdev = pci_get_drvdata(dev);
	
#ifdef DEBUG_IO
	printq("w %08x = %08x\n", reg, data);
#endif
	writel(data,pdev->baseaddress + reg);
}

u32 plx_reg_read32(struct pci_dev *dev,u32 reg)
{ 
	struct plx_dev *pdev = pci_get_drvdata(dev);
	u32 data;

	data = readl(pdev->baseaddress + reg);
#ifdef DEBUG_IO
	printq("r %08x = %08x\n", reg, data);
#endif
	return data;
}

void PLX9056_WriteConfigReg ( struct plx_dev * pdev, u32 reg, u32 val, int mode ) {
	switch (mode) {
	case SIZE_8BITS:
		writeb(val, pdev->iobase + reg);
		break;
	case SIZE_16BITS:
		writew(val, pdev->iobase + reg);
		break;
	case SIZE_32BITS:
		writel(val, pdev->iobase + reg);		
		break;
	default:
		printq("BUG in %s, reg is %08x\n\n", __FUNCTION__, reg);		
	}
}

u32 PLX9056_ReadConfigReg ( struct plx_dev * pdev, u32 reg, int mode ) {
	u32 r = 0;

	switch (mode) {
	case SIZE_8BITS:
		r = readb(pdev->iobase + reg);
		break;
	case SIZE_16BITS:
		r = readw(pdev->iobase + reg);
		break;
	case SIZE_32BITS:
		r = readl(pdev->iobase + reg);		
		break;
	default:
		printq("BUG in %s, reg is %08x\n\n", __FUNCTION__, reg);		
	}
	return r;
}

static void * rvmalloc(signed long size)
{
	struct page *page;
	void * mem;
	unsigned long adr;

	mem=vmalloc_32(size);
	if (NULL == mem)
		printq(KERN_INFO "bttv: vmalloc_32(%ld) failed\n",size);
	else {
		/* Clear the ram out, no junk to the user */
		memset(mem, 0, size);
	        adr=(unsigned long) mem;
		while (size > 0) {
			page = vmalloc_to_page((void *)adr);
			mem_map_reserve(page);
			adr+=PAGE_SIZE;
			size-=PAGE_SIZE;
		}
	}
	return mem;
}

static void rvfree(void * mem, signed long size)
{
	struct page *page;
        unsigned long adr;
        
	if (mem) {
	        adr=(unsigned long) mem;
		while (size > 0) {
			page = vmalloc_to_page((void *)adr);
			mem_map_unreserve(page);
			adr+=PAGE_SIZE;
			size-=PAGE_SIZE;
		}
		vfree(mem);
	}
}

static int alloc_dma_memory_global(struct plx_dma *dma) {
	int i;

	dma->vbuffer = rvmalloc(vbuffer_pages*PAGE_SIZE);
	dma->vhead = dma->vtail = 0;
	dma->phead = dma->ptail = 0;
	dma->overflow = 0;

	dma->img = (struct plx_image *) kmalloc(sizeof(struct plx_image) * vbuffer_images, GFP_KERNEL);
	memset(dma->img, 0, sizeof(struct plx_image) * vbuffer_images);
	for (i=0; i<vbuffer_images; i++) {
		dma->img[i].dma = dma;
		dma->img[i].pg_dma = dma_alloc_coherent(&dma->plx->pci->dev, PAGE_SIZE,
							&dma->img[i].pg_dma_phys, GFP_KERNEL);
		if (!dma->img[i].pg_dma) {
			printq("%s: coherent malloc failed for %d\n", __FUNCTION__, i);
		}
		dma->img[i].buf_virt = (unsigned long * *) get_zeroed_page(GFP_KERNEL);
	}
	return 0;
}

int alloc_dma_memory_image(struct plx_image *img, int size) {
	int i, j, n, nfree;
	unsigned long nextd, last;
	struct page *page;
	struct plx_dma *dma = img->dma;

	n = size/PAGE_SIZE;
	last = size & (PAGE_SIZE-1);
	if (last)
		n++;

	if (dma->vhead >= dma->vtail)
		nfree = vbuffer_pages - (dma->vhead - dma->vtail) - 1;
	else 
		nfree = dma->vtail - dma->vhead - 1;
	if (n > nfree)
		return -1;
	img->n = n;
	j = img->first = dma->vhead;

	dma->vhead = (dma->vhead + n) % vbuffer_pages;

	pr_debug("%s: %d -> %d(%ld) -> --- allocating %d-%d\n", __FUNCTION__, size, n, last, j, dma->vhead-1);

	if (img->status != IMAGE_IDLE) 
		printq("BUG @ %s %d\n", __FUNCTION__, __LINE__);
	img->status = IMAGE_DMA;
	
	nextd = img->pg_dma_phys;
	for (i=0 ; i<n; i++) {
		nextd += sizeof(struct plx_dma_desc);
		page = vmalloc_to_page((void *) ((unsigned long) dma->vbuffer) + j * PAGE_SIZE);
		img->pg_dma[i].BufferPhysicalAddress = dma_map_page(&dma->plx->pci->dev,
								    page, 0, PAGE_SIZE,
								    DMA_FROM_DEVICE);

		img->pg_dma[i].BufferLocalAddress = 0x80 + 4 * 1;

		if (i == (n-1) && last)
			img->pg_dma[i].BufferTransferSize = last;
		else
			img->pg_dma[i].BufferTransferSize = PAGE_SIZE;
			

		img->pg_dma[i].NextDescriptorPhysicalAddress = nextd |
			PLX_DESC_PTR_LOCATION_PCI | PLX_DESC_PTR_LCL_TO_PCI;

		j++;
		if (j >= vbuffer_pages)
		  j = 0;
	}
	img->pg_dma[i].BufferPhysicalAddress = 0;
	img->pg_dma[i].BufferLocalAddress = 0x80 + 4 * 1;
	img->pg_dma[i].BufferTransferSize = 0;
	img->pg_dma[i].NextDescriptorPhysicalAddress = PLX_DESC_PTR_END_OF_CHAIN | 
		PLX_DESC_PTR_LOCATION_PCI | PLX_DESC_PTR_LCL_TO_PCI;
	
	return 0;
}

int map_dma_memory_image(struct plx_image *img, int start, int size) {
	int i, j, n;
	unsigned long nextd, last;
	struct page *page;
	struct plx_dma *dma = img->dma;

	n = size/PAGE_SIZE;
	last = size & (PAGE_SIZE-1);
	if (last)
		n++;

	img->n = n;
	j = img->first = start;

	pr_debug("%s: %d -> %d(%ld) -> --- mapping %d-%d\n", __FUNCTION__, size, n, last, j, (j+n-1) % vbuffer_pages);

	if (img->status != IMAGE_USER) 
		printq("BUG @ %s %d\n", __FUNCTION__, __LINE__);
	img->status = IMAGE_DMA;
	
	nextd = img->pg_dma_phys;
	for (i=0 ; i<n; i++) {
		nextd += sizeof(struct plx_dma_desc);
		//pr_debug("%d(%d) -> %08lx\n", i, j, ((unsigned long) dma->vbuffer) + j * PAGE_SIZE);
		//pr_debug("data %08x\n", * (unsigned int *) (((unsigned long) dma->vbuffer) + j * PAGE_SIZE));
		page = vmalloc_to_page((void *) ((unsigned long) dma->vbuffer) + j * PAGE_SIZE);
		img->pg_dma[i].BufferPhysicalAddress = dma_map_page(&dma->plx->pci->dev,
								    page, 0, PAGE_SIZE,
								    DMA_TO_DEVICE);

		img->pg_dma[i].BufferLocalAddress = 0x80 + 4 * 1;

		if (i == (n-1) && last)
			img->pg_dma[i].BufferTransferSize = last;
		else
			img->pg_dma[i].BufferTransferSize = PAGE_SIZE;
			

		img->pg_dma[i].NextDescriptorPhysicalAddress = nextd |
			PLX_DESC_PTR_LOCATION_PCI;

		j++;
		if (j >= vbuffer_pages)
		  j = 0;
	}
	img->pg_dma[i].BufferPhysicalAddress = 0;
	img->pg_dma[i].BufferLocalAddress = 0x80 + 4 * 1;
	img->pg_dma[i].BufferTransferSize = 0;
	img->pg_dma[i].NextDescriptorPhysicalAddress = PLX_DESC_PTR_END_OF_CHAIN | 
		PLX_DESC_PTR_LOCATION_PCI;
	
	return 0;
}

void unmap_dma_memory_image(struct plx_image *img, int first, int n) {
	int i;
	struct plx_dma *dma = img->dma;

	if (img->status != IMAGE_DMA) 
		printq("BUG @ %s %d\n", __FUNCTION__, __LINE__);
	
	pr_debug("unmapping %d-%d\n", first, first+n-1);
	for(i=0; i<n; i++) {
		dma_unmap_page(&dma->plx->pci->dev,
			       img->pg_dma[i].BufferPhysicalAddress,
			       PAGE_SIZE, DMA_TO_DEVICE);
	}
	img->status = IMAGE_IDLE;
}

static void user_dma_memory(struct plx_image *img) {
	int i;
	struct plx_dma *dma = img->dma;

	if (img->status != IMAGE_DMA_DONE) 
		printq("BUG @ %s %d\n", __FUNCTION__, __LINE__);
	for(i=0; i<img->n; i++) {
		dma_unmap_page(&dma->plx->pci->dev,
			       img->pg_dma[i].BufferPhysicalAddress,
			       PAGE_SIZE, DMA_FROM_DEVICE);
	}
	img->status = IMAGE_USER;
}

static void free_dma_memory(struct plx_image *img) {
	struct plx_dma *dma = img->dma;
	u32 first;

	if (img->first != dma->vtail) {
		printq("BUG @ %s %d: something very wrong is going on with buffer mamagement!\n",
		       __FUNCTION__, __LINE__);
	}
	if (img->status != IMAGE_USER) 
		printq("BUG @ %s %d\n", __FUNCTION__, __LINE__);
	first = dma->vtail;
	dma->vtail = (dma->vtail + img->n) % vbuffer_pages;
	pr_debug("freeing %d-%d\n", first, dma->vtail-1);
	img->status = IMAGE_IDLE;
	
}

void free_dma_memory_head(struct plx_image *img) {
	struct plx_dma *dma = img->dma;
	u32 head, ohead;

	if (img->status != IMAGE_DMA_DONE) 
		printq("BUG @ %s %d\n", __FUNCTION__, __LINE__);

	user_dma_memory(img);

	ohead = head = dma->vhead;
	if (head < img->n)
		head += vbuffer_pages;
	head -= img->n;
	dma->vhead = head;
	pr_debug("head freeing %d-%d\n", head, ohead-1);
	img->status = IMAGE_IDLE;	
}


static void free_global_memory(struct plx_dma *dma) {
	int i;

	rvfree(dma->vbuffer, vbuffer_pages*PAGE_SIZE);
	for(i=0; i<vbuffer_images; i++) {
		dma_free_coherent(&dma->plx->pci->dev, PAGE_SIZE,
				  dma->img[i].pg_dma, dma->img[i].pg_dma_phys);
		free_page((unsigned long) dma->img[i].buf_virt);
	}
	kfree(dma->img);
}


void start_dma(struct plx_dma *dma, struct plx_image *img, int ch, unsigned long force) {
	struct plx_dev *bSlave;
	unsigned long	FirstDescriptorPhyAddr;
	unsigned long  DMAMode =
		PLX_DMA_MODE_LCLBUS32     	|
		PLX_DMA_MODE_SG_ENA       	|
		PLX_DMA_MODE_INT_DONE_ENA 	|
		PLX_DMA_MODE_CONST_LCL_ADDR|
		PLX_DMA_MODE_CLEAR_COUNT |
	  		PLX_DMA_MODE_DEMAND       	|
		PLX_DMA_MODE_INT_PCI;

	DMAMode |= PLX_DMA_MODE_TA_RDY_ENA;

	bSlave = dma->plx;
	if (force) {
		FirstDescriptorPhyAddr =  force;
	}
	else {
		FirstDescriptorPhyAddr =  img->pg_dma_phys;
	}
	FirstDescriptorPhyAddr =  FirstDescriptorPhyAddr |
		PLX_DESC_PTR_LOCATION_PCI | (ch==1 ? 0 : PLX_DESC_PTR_LCL_TO_PCI) ;
	
	switch (ch)
	{
	case 0:
		PLX9056_WriteConfigReg ( bSlave, PLX_ADDR_DMA0_MODE, DMAMode, SIZE_32BITS );
		PLX9056_WriteConfigReg ( bSlave, PLX_ADDR_DMA0_XFR_SIZE, 0, SIZE_32BITS );
		PLX9056_WriteConfigReg ( bSlave, PLX_ADDR_DMA0_DESC_POINTER, FirstDescriptorPhyAddr, SIZE_32BITS);
		PLX9056_WriteConfigReg ( bSlave, PLX_ADDR_DMA0_LCL_ADDR, 0x80 + 1 * 4, SIZE_32BITS);
		PLX9056_WriteConfigReg ( bSlave, PLX_ADDR_DMA0_CMDSTS, PLX_CMDSTS_ENABLE, SIZE_8BITS );
		PLX9056_WriteConfigReg ( bSlave, PLX_ADDR_DMA0_CMDSTS, PLX_CMDSTS_ENABLE | PLX_CMDSTS_START, SIZE_8BITS );
		break;
		
	case 1:
		PLX9056_WriteConfigReg ( bSlave, PLX_ADDR_DMA1_MODE, DMAMode, SIZE_32BITS );
		PLX9056_WriteConfigReg ( bSlave, PLX_ADDR_DMA1_XFR_SIZE, 0, SIZE_32BITS );
		PLX9056_WriteConfigReg ( bSlave, PLX_ADDR_DMA1_DESC_POINTER, FirstDescriptorPhyAddr, SIZE_32BITS );
		PLX9056_WriteConfigReg ( bSlave, PLX_ADDR_DMA1_LCL_ADDR, 0, SIZE_32BITS);
		PLX9056_WriteConfigReg ( bSlave, PLX_ADDR_DMA1_CMDSTS, PLX_CMDSTS_ENABLE, SIZE_8BITS );
		PLX9056_WriteConfigReg ( bSlave, PLX_ADDR_DMA1_CMDSTS, PLX_CMDSTS_ENABLE | PLX_CMDSTS_START, SIZE_8BITS );
		break;
	}	

	if (force)
		dma->is_dma_hdr = 1;
	else
		dma->is_dma = 1;
}

void stop_dma(struct plx_dma *dma, int DMAChannel) {
	struct plx_dev *bSlave = dma->plx;
	unsigned char DMACmdSts;
	unsigned long IntCtrlStat;
	unsigned long j;

	DMACmdSts = (unsigned char)PLX9056_ReadConfigReg( bSlave,
						DMAChannel ? PLX_ADDR_DMA1_CMDSTS : PLX_ADDR_DMA0_CMDSTS,
							sizeof(unsigned char) );

	if ( ( DMACmdSts & PLX_CMDSTS_ENABLE ) &&
			(!( DMACmdSts & PLX_CMDSTS_DONE ) ) )
	{
		//Mask DMA interrupt
		IntCtrlStat = PLX9056_ReadConfigReg( bSlave, PLX_ADDR_INTCTRLSTS, SIZE_32BITS);
		IntCtrlStat &= (DMAChannel ? (~PLX_INTCTRLSTS_DMA1_INT_ENA) : (~PLX_INTCTRLSTS_DMA0_INT_ENA));
		PLX9056_WriteConfigReg( bSlave, PLX_ADDR_INTCTRLSTS, IntCtrlStat, SIZE_32BITS );

		//Stop and abort DMA
		PLX9056_WriteConfigReg( bSlave, DMAChannel ? PLX_ADDR_DMA1_CMDSTS : PLX_ADDR_DMA0_CMDSTS,
				PLX_CMDSTS_ABORT, SIZE_8BITS);


		//Wait for completion
		j = jiffies + HZ;
		do {
			DMACmdSts = (unsigned char)PLX9056_ReadConfigReg( bSlave, DMAChannel ? PLX_ADDR_DMA1_CMDSTS : PLX_ADDR_DMA0_CMDSTS,
							SIZE_8BITS);
			schedule();
		} while (!(DMACmdSts & PLX_CMDSTS_DONE) && time_before(jiffies, j));
		if (!(DMACmdSts & PLX_CMDSTS_DONE)) {
			printq("%s error: DMACmdSts is %08x\n", __FUNCTION__, DMACmdSts);
		}

	}

	//Clear DMA interrupts

	PLX9056_WriteConfigReg( bSlave, DMAChannel ? PLX_ADDR_DMA1_CMDSTS : PLX_ADDR_DMA0_CMDSTS,
				 PLX_CMDSTS_CLRINT, SIZE_8BITS );

	dma->is_dma = 0;
}

void smup_start(struct plx_dma *dma) {
	dma->smup_desc[0].BufferPhysicalAddress = dma->smup_hdr_phys;
	dma->smup_desc[0].BufferLocalAddress = 0x80 + 4 * 1;
	dma->smup_desc[0].BufferTransferSize = sizeof(ADV202_COMPRESSED_HEADER);
	dma->smup_desc[0].NextDescriptorPhysicalAddress = (dma->smup_desc_phys + sizeof(struct plx_dma_desc)) |
	  PLX_DESC_PTR_LOCATION_PCI | PLX_DESC_PTR_LCL_TO_PCI;
	dma->smup_desc[1].BufferPhysicalAddress = 0;
	dma->smup_desc[1].BufferLocalAddress = 0x80 + 4 * 1;
	dma->smup_desc[1].BufferTransferSize = 0;
	dma->smup_desc[1].NextDescriptorPhysicalAddress = PLX_DESC_PTR_END_OF_CHAIN | 
		PLX_DESC_PTR_LOCATION_PCI | PLX_DESC_PTR_LCL_TO_PCI;
	start_dma(dma, NULL, 0, dma->smup_desc_phys);
}

static void init_dma(struct plx_dma *dma) {
	struct plx_dev *bSlave = dma->plx;

	pr_debug("%s %d\n", __FUNCTION__, __LINE__);
	// enable local address space 0, offset 0
	PLX9056_WriteConfigReg( bSlave, PLX_ADDR_LAS0BA, PLX_LADR0_ENABLE,SIZE_32BITS );
	// disable local burst and prefetch
	PLX9056_WriteConfigReg( bSlave, PLX_ADDR_LBRD0, PLX_REGSTARTUP_LBRD0,SIZE_32BITS );
	// set DMA endianess for PLX
	PLX9056_WriteConfigReg( bSlave, PLX_ADDR_ENDIAN, PLX_REGSTARTUP_ENDIAN,
				SIZE_8BITS );
	// set DMA arbitration mode
	PLX9056_WriteConfigReg( bSlave, PLX_ADDR_DMA_MODE_ARB, PLX_REGSTARTUP_DMA_MODE_ARB, SIZE_32BITS );
	// Disable interrupts
	PLX9056_WriteConfigReg( bSlave, PLX_ADDR_INTCTRLSTS,0, SIZE_32BITS );
	// Clear possibly pending interrupt flags
	PLX9056_WriteConfigReg( bSlave, PLX_ADDR_INTCTRLSTS,
				PLX_INTCTRLSTS_PM_INT_ACTIVE|PLX_INTCTRLSTS_PE_INT_ACTIVE, SIZE_32BITS );	

#if 0
	/* dump registers */
	{
	  int i;

	  for(i=0; i < 0x200; i+=4) {
	    printk("plx9056 dump %08x=%08x\n", i, PLX9056_ReadConfigReg ( bSlave, i, SIZE_32BITS ));
	    schedule();
	  }
	}
#endif
}

/* file operations */

void plx_vma_open(struct vm_area_struct *vma)
{
	pr_debug("VMA open, virt %lx, phys %lx\n",
		 vma->vm_start, vma->vm_pgoff << PAGE_SHIFT);
}

void plx_vma_close(struct vm_area_struct *vma)
{
	pr_debug("VMA close.\n");
}

struct page *plx_vma_nopage(struct vm_area_struct *vma,
                unsigned long address, int *type)
{
	struct plx_dma *pdev = (struct plx_dma *) vma->vm_private_data;
	struct page *pageptr;
	unsigned long offset = vma->vm_pgoff << PAGE_SHIFT;
	unsigned long virtaddr = address - vma->vm_start + offset;
	unsigned long pageframe;

	//pr_debug("nopage %08lx (from %08lx) ", virtaddr, address);
	while(virtaddr >= vbuffer_pages*PAGE_SIZE)
		virtaddr -= vbuffer_pages*PAGE_SIZE;
	//pr_debug("-> %08lx\n", (unsigned long) &pdev->vbuffer[virtaddr]);

	pageframe = kvirt_to_pa((unsigned long) &pdev->vbuffer[virtaddr]) >> PAGE_SHIFT;
	if (!pfn_valid(pageframe))
		return NOPAGE_SIGBUS;
	pageptr = pfn_to_page(pageframe);
	get_page(pageptr);
	if (type)
		*type = VM_FAULT_MINOR;
	return pageptr;
}

static struct vm_operations_struct plx_nopage_vm_ops = {
	.open =   plx_vma_open,
	.close =  plx_vma_close,
	.nopage = plx_vma_nopage,
};

static int plx_mmap(struct file *filp, struct vm_area_struct *vma)
{
	unsigned long offset = vma->vm_pgoff << PAGE_SHIFT;

	if (offset >= __pa(high_memory) || (filp->f_flags & O_SYNC))
		vma->vm_flags |= VM_IO;
	vma->vm_flags |= VM_RESERVED;

	vma->vm_ops = &plx_nopage_vm_ops;
	vma->vm_private_data = filp->private_data;
	plx_vma_open(vma);
	return 0;
}



static ssize_t plx_read(struct file *filp, char __user *buf, size_t count,
                loff_t *f_pos)
{
	struct plx_dma *pdev = (struct plx_dma *) filp->private_data;
	struct plx_image *img;
	int ret;
	unsigned long timeout = HZ + /* CHRI, think about this timeout */
		HZ * pdev->skip / (pdev->is_ntsc ? 60 : 50) * (pdev->is_interlaced ? 2 : 1);

	//dev_dbg(&dev->dev, "%s %d\n", __FUNCTION__, __LINE__);

	if ((pdev->opened & PLX_OPENED_IN) == PLX_OPENED_IN) {
		if (count < sizeof(struct ethjpg2k_linux_hdr)) {
			dev_dbg(&pdev->plx->pci->dev, "size\n");		
			return -EINVAL;
		}
	  
		if (pdev->phead == pdev->ptail) {
			if (pdev->plx->mode == PLX_HIPI_ENCODE || pdev->plx->mode == PLX_HIPI_DECODE)
				timeout = 10*HZ;
			ret = wait_event_interruptible_timeout(pdev->pready, (pdev->phead != pdev->ptail),
							       timeout);
			if (ret == 0) {
				printq("timeout waiting for image\n");
				return -ETIMEDOUT;
			}
			else {
				if (ret == -ERESTARTSYS)
					return ret;
			}
		}


		img = &pdev->img[pdev->ptail];
		pr_debug("reading slot %d: %08lx %8ld %08lx %ld\n", pdev->ptail, img->hdr.field, 
			 img->hdr.index, img->hdr.Info_u.info, img->hdr.size);
		user_dma_memory(img);

		img->hdr.start = img->first * PAGE_SIZE;
		img->hdr.overflow = pdev->overflow;
		pdev->overflow = 0;
		img->hdr.ticket = pdev->ptail;

		if (copy_to_user(buf, &img->hdr, sizeof(struct ethjpg2k_linux_hdr))) 
			return -EFAULT;			      
		return sizeof(struct ethjpg2k_linux_hdr);
	}
	else if ((pdev->opened & PLX_OPENED_OUT) == PLX_OPENED_OUT) {
		struct ethjpg2k_linux_decode d;
		u32 vhead, vtail;
		
		if (count < sizeof(struct ethjpg2k_linux_decode)) {
			dev_dbg(&pdev->plx->pci->dev, "decode size\n");		
			return -EINVAL;
		}
	  
		if ((pdev->phead + 1) % vbuffer_images == pdev->ptail) {
			if (pdev->plx->mode == PLX_HIPI_ENCODE || pdev->plx->mode == PLX_HIPI_DECODE)
				timeout = 10*HZ;
			ret = wait_event_interruptible_timeout(pdev->pready,
							       ((pdev->phead + 1) % vbuffer_images != pdev->ptail),
							       timeout);
			if (ret == 0)
				return -ETIMEDOUT;
			else {
			  if (ret == -ERESTARTSYS)
			    return ret;
			}
		}
		vhead = pdev->vhead;
		vtail = pdev->vtail;
		if (vhead >= vtail)
			d.size = vbuffer_pages - (vhead - vtail) - 1;
		else 
			d.size = vtail - vhead - 1;
		d.size *= PAGE_SIZE;
		d.start = vhead * PAGE_SIZE;
		dev_dbg(&pdev->plx->pci->dev, "free from %ld and %ld bytes\n", d.start, d.size);		
		if (copy_to_user(buf, &d, sizeof(d))) 
			return -EFAULT;			      
		return sizeof(d);
	}
	else {
		dev_dbg(&pdev->plx->pci->dev, "bad direction\n");
	}
	
	return -EBADFD;
}

static ssize_t plx_write(struct file *filp, const char __user *buf, size_t count,
                loff_t *f_pos)
{
	struct plx_dma *pdev = (struct plx_dma *) filp->private_data;
	struct plx_image *img;
	unsigned int ticket;

	if ((pdev->opened & PLX_OPENED_IN) == PLX_OPENED_IN) {

		if (count != sizeof(unsigned int)) {
			dev_dbg(&pdev->plx->pci->dev, "wsize\n");		
			return -EINVAL;
		}

		if (copy_from_user(&ticket, buf, sizeof(unsigned int)))
			return -EFAULT;

		pr_debug("write for ticket %d\n", ticket);

		if (pdev->ptail != ticket) {
			printq("BUG @ %s %d: something very wrong is going on with buffer mamagement!\n",
			       __FUNCTION__, __LINE__);
		}
	
		img = &pdev->img[ticket];
		free_dma_memory(img);
		pdev->ptail++;
		if (pdev->ptail == vbuffer_images)
			pdev->ptail = 0;

		return count;
	}
	else if ((pdev->opened & PLX_OPENED_OUT) == PLX_OPENED_OUT){
		int np;

		if (count != sizeof(struct ethjpg2k_linux_decode)) {
			dev_dbg(&pdev->plx->pci->dev, "wwsize\n");		
			return -EINVAL;
		}
		img = &pdev->img[pdev->phead];
		if (img->status != IMAGE_IDLE) {
			printk("BUG @ %s %d\n", __FUNCTION__, __LINE__);
			return -EBUSY;
		}
		if (copy_from_user(&img->dec, buf, sizeof(struct ethjpg2k_linux_decode)))
			return -EFAULT;
		img->dec.start /= PAGE_SIZE;
		np = img->dec.size / PAGE_SIZE;
		if (img->dec.size & (PAGE_SIZE-1))
			np++;
			
		dev_dbg(&pdev->plx->pci->dev, "for image %d reserving %d-%d\n", 
			pdev->phead, pdev->vhead,  (pdev->vhead + np -1) % vbuffer_pages);		
		img->status = IMAGE_USER;
		pdev->vhead = (pdev->vhead + np) % vbuffer_pages;
		pdev->phead = (pdev->phead + 1 ) % vbuffer_images;
		tasklet_schedule(&pdev->task);		
		return count;
	}
	else {
		dev_dbg(&pdev->plx->pci->dev, "bad direction\n");
	}
	return -EBADFD;
}

static int plx_release(struct inode *inode, struct file *filp)
{
	struct plx_dma *pdev = (struct plx_dma *) filp->private_data;
	struct plx_dev *bSlave = pdev->plx;
	dev_dbg(&pdev->plx->pci->dev, "%s %d\n", __FUNCTION__, __LINE__);

	if (pdev->opened & PLX_RUNNING) {
		stop_encdec(pdev);
	}

	free_global_memory(pdev);

	bSlave->mode = PLX_NONE;

	if ((pdev->opened & PLX_OPENED_IN)  == PLX_OPENED_IN) {
		dma_free_coherent(&pdev->plx->pci->dev,
				  2*sizeof(struct plx_dma_desc),
				  pdev->smup_desc, 
				  pdev->smup_desc_phys);

		dma_free_coherent(&pdev->plx->pci->dev,
				  sizeof(ADV202_COMPRESSED_HEADER),
				  pdev->smup_hdr, 
				  pdev->smup_hdr_phys);

	}
	if ((pdev->opened & PLX_OPENED_OUT)  == PLX_OPENED_OUT) {
		tasklet_disable(&pdev->task);
	}
	
	pdev->opened = PLX_CLOSED;
	return 0;
}

static int plx_open(struct inode *inode, struct file *filp)
{
	struct plx_dma *pdev;
	struct plx_dev *bSlave;

	pr_debug("%s %d %d\n", __FUNCTION__, __LINE__, iminor(inode));		

	pdev = container_of(inode->i_cdev, struct plx_dma, cdev);
	bSlave = pdev->plx;

	if (pdev->opened != PLX_CLOSED)
		return -EEXIST;

	/* 
	   mode              minor 0 (enc)                            minor 1 (dec)
	   PLX_ENCODE        PLX_OPENED_IN (outputs encoded data)     none
	   PLX_DECODE        none                                     PLX_OPENED_OUT (input encoded data)
	   PLX_HIPI_ENCODE   PLX_OPENED_IN (outputs encoded data)     PLX_OPENED_OUT (input raw data)
	   PLX_HIPI_DECODE   PLX_OPENED_IN (outputs raw data)         PLX_OPENED_OUT (input encode data)
	   PLX_LOOP          PLX_OPENED_IN (outputs raw data)         PLX_OPENED_OUT (input raw data)

	   DMA: channel 0 tied to minor 0 where data goes from ADV202 to CPU so plx_dev->dma[0]
	        channel 1 tied to minor 1 where data goes from CPU to ADV202 so plx_dev->dma[1]
	*/

	if (iminor(inode) & 1) {
		pdev->opened = PLX_OPENED_OUT;
		bSlave->mode = PLX_DECODE;
	}
	else {
		pdev->opened = PLX_OPENED_IN;
		bSlave->mode = PLX_ENCODE;
	}
	pr_debug("%s pdev %p opened %08x mode %08x\n", __FUNCTION__,
		 pdev, pdev->opened, bSlave->mode);
	alloc_dma_memory_global(pdev);	

	if ((pdev->opened & PLX_OPENED_IN)  == PLX_OPENED_IN) {
		pdev->smup_desc = dma_alloc_coherent(&pdev->plx->pci->dev, 
						   2*sizeof(struct plx_dma_desc),
						   &pdev->smup_desc_phys, 
						   GFP_KERNEL);
		pdev->smup_hdr = dma_alloc_coherent(&pdev->plx->pci->dev, 
						    sizeof(ADV202_COMPRESSED_HEADER),
						    &pdev->smup_hdr_phys, 
						    GFP_KERNEL);
	}
	if ((pdev->opened & PLX_OPENED_OUT)  == PLX_OPENED_OUT) {
		tasklet_enable(&pdev->task);
	}

	pdev->phead = pdev->ptail = pdev->is_dma = 0;
	pdev->is_dma_hdr = 0;
	filp->private_data = pdev; 	

	return 0;
}

struct file_operations plx_fops = {
	.owner =    THIS_MODULE,
	.read =     plx_read,
	.write =    plx_write,
	.open =     plx_open,
	.release =  plx_release,
	.ioctl =    plx_ioctl,
	.mmap =	    plx_mmap,
};

static int plx_setup_cdev(struct cdev  *cdev, int index)
{
	int err = 0, devno = MKDEV(plx_major, index);
    
	cdev_init(cdev, &plx_fops);
	cdev->owner = THIS_MODULE;
	cdev->ops = &plx_fops;
	err = cdev_add (cdev, devno, 1);
	/* Fail gracefully if need be */
	if (err)
		printq("Error %d adding index %d", err, index);
	return err;
}

static int __devinit
plx_pci_probe (struct pci_dev *dev, const struct pci_device_id *id)
{
	struct plx_dev  *pdev;
	int status = 1;
	dev_t devn = 0;
	int i;
    
	dev_dbg(&dev->dev,"%s %d\n", __FUNCTION__, __LINE__);

	if (pci_enable_device(dev) < 0){
		printq("failed in enabing the device\n");
		return -ENODEV;
	}
	if (!dev->irq) {
		printq("found plx device with no IRQ assigned."); 
		printq("check BIOS settings!");
		return -ENODEV;
	}

	pdev = kmalloc(sizeof(*pdev), GFP_KERNEL);
	memset(pdev,0,sizeof(*pdev));
	init_waitqueue_head(&pdev->dma[0].pready);
	init_waitqueue_head(&pdev->dma[1].pready);
	pdev->dma[0].plx = pdev;
	pdev->dma[1].plx = pdev;
	pdev->pci = dev;
	sprintf(pdev->name, "plx%d", plx_minor);
	
	/* Grab the PLX PCI mem maped port start address we need  */
	pdev->pci_io_base = pci_resource_start(dev, 0);
	dev_dbg(&dev->dev, "pci io base %08x, irq %d\n", pdev->pci_io_base, dev->irq);;

	pdev->iolength = pci_resource_len(dev, 0);
	dev_dbg(&dev->dev, "pci io length %d\n", pdev->iolength);
	if(!request_mem_region(pdev->pci_io_base, pdev->iolength, pdev->name)){
		printq("host controller already in use!\n");		
		status = -EBUSY;
		goto err0;
	}   

	pdev->iobase = ioremap_nocache(pdev->pci_io_base, pdev->iolength);
	if(!pdev->iobase){
		printq("can not map io memory to system memory\n");
		status = -ENOMEM;
		goto err1;
	}

	/* Grab the PLX PCI shared memory */
	pdev->pci_mem_phy0 = pci_resource_start (dev, 2);         
	dev_dbg(&dev->dev, "pci base address = %x\n", pdev->pci_mem_phy0);
	pdev->length = pci_resource_len(dev,2);

	if(check_mem_region(pdev->pci_mem_phy0, pdev->length)<0){
		printq("host controller already in use\n");
		status = -EBUSY;
		goto err2;
	}
	
	if(!request_mem_region(pdev->pci_mem_phy0, pdev->length, pdev->name)){
		printq("host controller already in use\n");
		status =  -EBUSY;
		goto err2;
	}

	/*map available memory*/
	pdev->baseaddress = ioremap_nocache(pdev->pci_mem_phy0,pdev->length);
	if(!pdev->baseaddress){
		printq("memory map problem\n");
		status =  -ENOMEM;
		goto err3;
	} 
	pethjpg2k->PLX9056_BAR2[plx_minor/2] = (unsigned long) pdev->baseaddress;
	pethjpg2k->plx[plx_minor/2] = pdev;

	pci_set_drvdata (dev, pdev);
	plx_request_irq(pdev->pci, pdev->pci);
	init_dma(&pdev->dma[0]);

	pci_set_master(dev);	/* TODO: check this */
	
	if (plx_major) {
		devn = MKDEV(plx_major, plx_minor);
		status = register_chrdev_region(devn, 2, "plx");
	} else {
		status = alloc_chrdev_region(&devn, plx_minor, 2,
					     "plx");
		plx_major = MAJOR(devn);
	}
	if (status < 0) {
		printq("plx: can't get major %d\n", plx_major);
		goto err4;
	}


	for(i=0;i<2;i++) {
		dev_dbg(&dev->dev, "registering minor %d\n", plx_minor);
		pdev->dma[i].minor = plx_minor;
		plx_setup_cdev(&pdev->dma[i].cdev, plx_minor);
		tasklet_init(&pdev->dma[i].task, decode_tasklet, (unsigned long) &pdev->dma[i]);
		tasklet_disable(&pdev->dma[i].task);
		dev_dbg(&dev->dev, "%p %p\n", pdev, &pdev->dma[i].cdev);
		plx_minor ++; 
	}

	pethjpg2k->Bridges++;

	return 1;

        // err5:
	//unregister_chrdev_region(devn, 2);

 err4:
	iounmap(pdev->baseaddress);
 err3:    
	release_mem_region(pdev->pci_mem_phy0, pdev->length);
 err2:
	iounmap(pdev->iobase);
 err1:
	release_mem_region(pdev->pci_io_base, pdev->iolength);
 err0:
	kfree(pdev);
	return status;
}

static void __devexit
plx_pci_remove (struct pci_dev *dev)
{
	struct plx_dev *pdev = pci_get_drvdata(dev);
	dev_t devn = MKDEV(plx_major, pdev->dma[0].minor);
	
	cdev_del(&pdev->dma[0].cdev);
	cdev_del(&pdev->dma[1].cdev);
	unregister_chrdev_region(devn, 2);

	plx_free_irq(dev, dev);

	release_mem_region(pdev->pci_io_base, pdev->iolength);      
	release_mem_region(pdev->pci_mem_phy0, pdev->length);
	iounmap(pdev->baseaddress);
	iounmap(pdev->iobase);

	kfree(pdev);

	return;
}


static const struct pci_device_id __devinitdata plx_pci_ids [] = {
	{
		/* handle PCI BRIDGE  manufactured by PLX */
		class:          ((PCI_CLASS_BRIDGE_OTHER << 8) | (0x06 << 16)),
#ifdef ANALOG_DEMO
		class_mask:     0,
		vendor:         0x11d4,
		device:         0x0202,
#else
		class_mask:     ~0,
		vendor:         0x10B5,
		device:         0x9056,
#endif
		subvendor:      PCI_ANY_ID,
		subdevice:      PCI_ANY_ID,
	}, 
	{ /* end: all zeroes */ }
};

MODULE_DEVICE_TABLE (pci, plx_pci_ids);

/* Pci driver interface functions */
static struct pci_driver plx_pci_driver = {
	name:           "plx9056",
	id_table:       &plx_pci_ids[0],
	probe:         plx_pci_probe,
	remove:        plx_pci_remove,
};

static int __init plx_module_init(void) 
{
	int result = 0;
	pr_debug("%s: %d\n",__FUNCTION__, __LINE__);

	pethjpg2k = kmalloc(sizeof(*pethjpg2k), GFP_KERNEL);
	if (!pethjpg2k) {
		printq("kmallof failed at %s %d\n", __FUNCTION__, __LINE__);
		return -ENOMEM;
	}
	memset(pethjpg2k, 0, sizeof(*pethjpg2k));
	spin_lock_init(&pethjpg2k->csMotionStatus);

	if((result = pci_module_init(&plx_pci_driver)) < 0) {
		printq("PCI Iinitialization Fail(error = %d)\n",result);
		return result;
	}
	else
		pr_debug("%s PCI Initialization Success \n",plx_pci_driver.name);
	return result;
}

static void __exit plx_module_cleanup (void) 
{
	pr_debug("%s: %d\n",__FUNCTION__, __LINE__);

	pci_unregister_driver (&plx_pci_driver);
	kfree(pethjpg2k);
} 


MODULE_AUTHOR ("<chri@evolware.org>");
MODULE_DESCRIPTION("Eurotech e0904 JPEG2000 card driver");
MODULE_LICENSE ("GPL");

module_init (plx_module_init);
module_exit (plx_module_cleanup);


