/*
  Copyright 2007 Eurotech S.P.A. (Luca Calligaris)
  Copyright 2007 Christian Pellegrin <chripell@gmail.com> EVOL S.R.L.
  
  This software is released under the GPL v2. Please see the file
  COPYING in the top-level directory.
*/

#ifndef _PLX9056_H_
#define _PLX9056_H_ 1

#include <linux/version.h>
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,19)
#include <linux/config.h>
#endif
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/cdev.h>
#include <linux/interrupt.h>

#include "ethjpg2k.h"
#include "adv202_hdr.h"

struct plx_dma_desc {
	unsigned long	BufferPhysicalAddress;
	unsigned long	BufferLocalAddress;
	unsigned long	BufferTransferSize;
	unsigned long	NextDescriptorPhysicalAddress;
};

struct plx_dev;
struct plx_dma;

struct plx_image {
	int status;
#define IMAGE_IDLE 0
#define IMAGE_DMA 1
#define IMAGE_DMA_DONE 2
#define IMAGE_READY 3
#define IMAGE_USER 4

	int first;		/* first page used */
	int n;			/* number of pages used */

	struct plx_dma_desc *pg_dma;                /* DMA's page table address*/
	dma_addr_t pg_dma_phys;      /* DMA's page table physical address*/

	unsigned long * *buf_virt; /* virtual address of pages of this buffer */

	struct plx_dma *dma;

	struct ethjpg2k_linux_hdr hdr; /* for encoding */
	struct ethjpg2k_linux_decode dec; /* for decoding */
};

struct plx_dma {
	/* -- generic */

	volatile int opened;
#define PLX_CLOSED 0
#define PLX_OPENED_IN 1
#define PLX_OPENED_OUT 2
#define PLX_RUNNING 4
	struct plx_dev *plx;
	int minor;
	struct cdev cdev;

	/* -- buffer */

	/* buffer page management */
	u8 *vbuffer;
	volatile u32 vhead;
	volatile u32 vtail;

	/* we are trasfering to vhead and reading to vtail */
	volatile u32 phead;
	volatile u32 ptail;

	/* images buffers */
	struct plx_image * img;	

	/* -- both encoding and decoding */

	int is_dma;		/* for consistency check */
	int is_dma_hdr;		/* for consistency check */
	
	wait_queue_head_t pready; /* for waking up read */

	/* -- encoding */

	/* overflow indicator */
	int overflow;

	dma_addr_t smup_desc_phys;
	struct plx_dma_desc *smup_desc;
	
	dma_addr_t smup_hdr_phys;
	ADV202_COMPRESSED_HEADER *smup_hdr;

	/* -- decoding */

	struct tasklet_struct task; /* starts decoding of a new image */

	/* -- parameters */

	int skip;		/* how many frames we skip in HW */
	int is_ntsc;		/* the norm for frame rate calculation */
	int is_interlaced;	/* if we do de-interlation */
	int coding_style;
};

struct plx_dev {
	char name[16];
	struct pci_dev *pci;
	int mode;
#define PLX_NONE 0
#define PLX_ENCODE 1
#define PLX_DECODE 2
#define PLX_HIPI_ENCODE 3
#define PLX_HIPI_DECODE 4
#define PLX_HIPI_LOOP 5

	void *baseaddress;
	u32 pci_mem_phy0;
	u32 length;

	void *iobase;
	u32 pci_io_base;
	u32 iolength;

	struct plx_dma dma[2];
	/* 
	   note: dma[0] is always input (adv202 encodes and send to
	   CPU), dma[1] is output (adv202 decodes and gets data from
	   CPU)
	*/

	int image_size;		/* for HIPI */
	int adv_started;
};


void plx_irqon(struct plx_dma *dma, u32 Mask);
void plx_irqoff(struct plx_dma *dma, u32 Mask);
#ifdef SMUP
void smup_start(struct plx_dma *dma);
#endif
void start_dma(struct plx_dma *dma, struct plx_image *img, int ch, unsigned long force);
void stop_dma(struct plx_dma *dma, int DMAChannel);
void PLX9056_WriteConfigReg ( struct plx_dev * pdev, u32 reg, u32 val, int mode );
u32 PLX9056_ReadConfigReg ( struct plx_dev * pdev, u32 reg, int mode );
int plx_ioctl(struct inode *inode, struct file *filp,
	      unsigned int cmd, unsigned long arg);
int plx_request_irq(struct pci_dev *dev, void *isr_data);
void plx_free_irq(struct pci_dev *dev, void *isr_data);
int alloc_dma_memory_image(struct plx_image *img, int size);
void stop_encdec(struct plx_dma * pdev);
void adv202_encode_work(void * pdev_);
void restarter_work(void * pdev_);
void free_dma_memory_head(struct plx_image *img);
int my_queue_work(struct workqueue_struct *queue, struct work_struct *work, struct plx_dma *pdev);
void handle_data_encode(struct plx_dma * dma, int force);
void decode_tasklet(unsigned long pdev_);
int map_dma_memory_image(struct plx_image *img, int start, int size);
void unmap_dma_memory_image(struct plx_image *img, int first, int n);
void smup_start(struct plx_dma *dma);
void smup_setup(struct plx_dma * pdev);

extern int quiet;
#define printq(arg...) \
do { \
if (quiet) \
pr_debug(arg); \
else \
printk(arg); \
} while (0)

#endif
