
/*
  Copyright 2007 Eurotech S.P.A. (Luca Calligaris)
  Copyright 2007 Christian Pellegrin <chripell@gmail.com> EVOL S.R.L.
  
  This software is released under the GPL v2. Please see the file
  COPYING in the top-level directory.
*/

#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <netinet/in.h>

#include "libe0904.h"

#define MAX_PATH 255

int is_ntsc = 0;

void usage(void) {
  printf("usage: test_decode [dev] [n sec] [jp2 images .....]\n");
  exit(1);
}

char firmware[MAX_PATH]="../firmware/decode_2_4_12.sea"; 
//char firmware[MAX_PATH]="../firmware/decode_2_3f.sea"; 
ETHJPG2K_FIRMWARE_DATA fwdata = {};
ETHJPG2K_FIRMWARE_PARAMS fwparams = {
  /* these are values suggested by Christine Bako*/
  .DecodeParams= {
    .VFORMAT = 0,		/* will be automatically filled */
    .UNI = 3,    
    .COD_STYLE = 2,		/* 1 is JC2, 2 is JP2 */
    .PICFG = 0x11,
  },
};

struct image_s {
  char *data;
  int size;
} *images;

int cur, n, c=0;
time_t finish_epoch;

void get_image(unsigned char * data, unsigned long len) {
#if 0
  struct timeval tv;

  gettimeofday(&tv, NULL);
  fprintf(stderr, "%d %ld.%ld\n", c, tv.tv_sec, tv.tv_usec);
#endif
  memcpy(data, images[cur].data, len);
  cur = (cur + 1) % n;
  c++;
}

static time_t first = 0;
int get_len (void) {
  time_t last;

  if (time(&last) > finish_epoch) {
    if (c>0) 
      printf("%f fps\n", ((float) c) / (last-first) );
    return -1;
  }
  if (!first)
    first = last;
  return images[cur].size;
}


int main(int argc, char *argv[]) {
  int i;
  char *dev;
  struct e0904_ch *e;
  AP5200_MODE_PARAMS p;
  AP5200_PLAYBACK pb;

  if (argc < 4) {
    usage();
  }
  dev = argv[1];

  if (is_ntsc)
    fwparams.DecodeParams.VFORMAT = 0;
  else
    fwparams.DecodeParams.VFORMAT = 1;

  n = argc - 3;
  images = calloc(n, sizeof(struct image_s));
  assert(images);
  for(i=0; i<n; i++) {
    char *fname = argv[i+3];
    FILE *f;
    int padded;


    f = fopen(fname, "r");
    if (!f) {
      fprintf(stderr, "Cannot open %s!\n", fname);
      exit(1);
    }
    fseek(f, 0, SEEK_END);
    images[i].size = ftell(f);
    assert(images[i].size);
    printf("image %d size %d\n", i, images[i].size);
    rewind(f);
    padded = images[i].size;
    images[i].data = calloc(1, padded);
    assert(images[i].data);
    fread(images[i].data, images[i].size, 1, f);
    fclose(f);
    images[i].size = padded;
  }

  e = simple_e0904_encoder(argv[1]);
  assert(e);
  assert(!simple_e0904_board_init(e, is_ntsc, firmware, &fwdata, &fwparams, 0));

  memset(&p, 0, sizeof(p));
  p.Mode = MODE_MON_FULL;
  p.MON_FULL.Channel = 0;
  simple_e0904_ioctl(e, IOCTL_ETHJPG2K_VIDEO_SET_MODE, (void *) &p);

  memset(&pb, 0, sizeof(pb));
  pb.dest_ch = 0;
  pb.enable = 1;
  pb.isNtsc = is_ntsc;     
  simple_e0904_ioctl(e, IOCLT_ETHJPG2K_VIDEO_PLAYBACK, (void *) &pb);

  finish_epoch = time(NULL) + strtoul(argv[2], NULL, 0);
  cur = 0;
  simple_e0904_decoder_run(e, get_len, get_image, 1);

  memset(&pb, 0, sizeof(pb));
  pb.enable = 0;
  simple_e0904_ioctl(e, IOCLT_ETHJPG2K_VIDEO_PLAYBACK, (void *) &pb);

  return 0;
}
