/*
  Copyright 2007 Eurotech S.P.A. (Luca Calligaris)
  Copyright 2007 Christian Pellegrin <chripell@gmail.com> EVOL S.R.L.
  
  This software is released under the GPL v2. Please see the file
  COPYING in the top-level directory.
*/

#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <pthread.h>
#include <unistd.h>

#include "libe0904.h"

#define MAX_PATH 255

/* are we running? */
volatile int running = 0;
volatile int ack_finished = 0;

/* the encoder ant it's thread*/
struct e0904_ch* enc = NULL;
pthread_t eth;

/* global configuration parameters */
int is_ntsc = 0;
int slave_path = 0;
char firmware[MAX_PATH]="../firmware/encode_2_9_2.sea"; 
ETHJPG2K_FIRMWARE_DATA fwdata = {};
ETHJPG2K_FIRMWARE_PARAMS fwparams = {
  /* these are values suggested by Christine Bako*/
  .EncodeParams= {
    .VFORMAT = 1,
    .XFORMLEV = 5,
    .UNI = 3,
    .RCTYPE = 2,
    .RCVAL = {0, 5, 0},
    .COD_STYLE = 2,
    .PICFG = 1,
  },
};
volatile int verbose;			/* 0 quiet, 1 just errors, 2 verbose */
volatile int every = 0;			/* if > 0 take a pic every every */
char every_prefix[MAX_PATH];		/* prefix to use */
volatile int shot = 0;			/* take just a shot */
char shot_name[MAX_PATH];	/* name of the shot */

/* running parameters */
volatile int dropped = 0;
volatile int frames = 0;
volatile int resets = 0;

/* callback that processes every captured frame */
static int cb(struct ethjpg2k_linux_hdr *hdr, unsigned char * data, unsigned long len) {
  char b[255];
  FILE *f;
  static int private_prog = 0;

  if (verbose == 2) {
    printf("%08d: %08lx %08ld %08lx %08ld -> %08ld ! %d T %d\n", frames, hdr->field, hdr->index, hdr->Info_u.info, 
	   hdr->size*4, hdr->start, hdr->overflow, hdr->ticket );
  }
  else if (verbose == 1) {
    if (hdr->index < private_prog) {
      printf("reset needed after %d frames\n", private_prog);
    }
    if (hdr->overflow > 0) {
      printf("dropped %d frames\n", hdr->overflow);
    }
  }  

  if (every > 0 && (frames % every == 0)) {
    sprintf(b, "%s%08d.jp2", every_prefix, frames);
    f = fopen(b, "w");
    assert(f);
    fwrite(data, len, 1, f);
    fclose(f);
  }

  if (shot) {
    f = fopen(shot_name, "w");
    assert(f);
    fwrite(data, len, 1, f);
    fclose(f);
    shot = 0;
  }

  if (hdr->index < private_prog) {
    resets++;
    private_prog = 0;
  }
  else
    private_prog ++;
  
  dropped += hdr->overflow;
  frames++;

  if (!running)
    return 1;
  return 0;
}

/* acquiring thread */
void * e0904_thread(void * priv) {
  
  pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
  running = 1;  
  ack_finished = 0;
  dropped = frames = resets = 0;
  do {
    assert(!simple_e0904_board_init(enc, is_ntsc, firmware, &fwdata, &fwparams, slave_path));
    simple_e0904_encoder_run(enc, cb);
  } while(running);
  ack_finished = 1;

  return NULL;
}

/* callback type for command */
typedef int (*cmd_cb) (int argc, char *argv[]);

/* commands */

int do_every(int argc, char *argv[]) {
  every = strtol(argv[1], NULL, 0);
  strcpy(every_prefix, argv[2]);
  return 0;
}

int do_fw(int argc, char *argv[]) {
  strcpy(firmware, argv[1]);
  return 0;
}

int do_help(int argc, char *argv[]);

int do_shot(int argc, char *argv[]) {
  if (!running) {
    printf("not running\n");
    return 1;
  }

  strcpy(shot_name, argv[1]);
  shot = 1;
  return 0;
}

int do_start(int argc, char *argv[]) {
  int res;

  if (running) {
    printf("already running\n");
    return 1;
  }

  res = pthread_create(&eth, NULL, e0904_thread, NULL);
  assert(res == 0);
  
  return 0;
}

int do_stat(int argc, char *argv[]) {
  printf("r %d f %d d %d r %d\n", running, frames, dropped, resets);
  return 0;
}

int do_stop(int argc, char *argv[]) {
  int res, i;
  void *thread_result;

  if (!running) {
    printf("not running\n");
    return 1;
  }

  running = 0;

#define LOOP 10
  for(i=0; i<LOOP; i++) {
    if (ack_finished)
      break;
    usleep(10 * 1000);
  }
  if (i==LOOP) {			/* brute force */
    simple_e0904_force_stop(enc);
    pthread_cancel(eth);
  }
#undef LOOP

  res = pthread_join(eth, &thread_result);

  return 0;
}

int do_verbose(int argc, char *argv[]) {
  verbose = strtol(argv[1], NULL, 0);
  return 0;
}

int do_quit(int argc, char *argv[]) {
  exit(0);
}

int do_wait(int argc, char *argv[]) {
  int target;
  
  if (!running) {
    printf("not running\n");
    return 1;
  }
  
  target = frames + strtol(argv[1], NULL, 0);
  while (frames < target)
    usleep(10 * 1000);
  return 0;
}

int do_quality(int argc, char *argv[]) {
  int val, i;  

  fwparams.EncodeParams.RCTYPE = strtol(argv[1], NULL, 0);
  val = strtol(argv[2], NULL, 0);
  for(i=2; i>=0; i--) {
    fwparams.EncodeParams.RCVAL[i] = val % 256;
    val /= 256;
  }
  return 0;
}

int do_skip(int argc, char *argv[]) {
  fwparams.EncodeParams.STALLPAR = strtol(argv[1], NULL, 0);
  return 0;
}

int do_norm(int argc, char *argv[]) {
  if (!strcmp(argv[1], "pal")) {
    is_ntsc = 0;
    fwparams.EncodeParams.VFORMAT = 1;    
  }
  else if (!strcmp(argv[1], "pali")) {
    is_ntsc = 0;
    fwparams.EncodeParams.VFORMAT = 9;    

    /* these parameters are from rev19 example Analog Board */

    fwparams.EncodeParams.VFORMAT = 9;    
    fwparams.EncodeParams.PREC = 1;    
    fwparams.EncodeParams.XFORMLEV = 5;    
    fwparams.EncodeParams.UNI = 3;    

    fwparams.EncodeParams.CBSIZE = 3;    
    fwparams.EncodeParams.WKERNEL_QUANT = 0;    
    fwparams.EncodeParams.STALLPAR = 0;    
    fwparams.EncodeParams.ATTRTYPE = 0;    

    fwparams.EncodeParams.RCTYPE = 0;    
    fwparams.EncodeParams.RCVAL[0] = 0;    
    fwparams.EncodeParams.RCVAL[1] = 2;    
    fwparams.EncodeParams.RCVAL[2] = 0;    

    fwparams.EncodeParams.J2KPROG = 0;    
    fwparams.EncodeParams.PICFG = 1;    
    fwparams.EncodeParams.QFACT = 0;    
    fwparams.EncodeParams.COD_STYLE = 2;    
  }
  else if (!strcmp(argv[1], "ntsc")) {
    is_ntsc = 1;
    fwparams.EncodeParams.VFORMAT = 0;
  }
  else if (!strcmp(argv[1], "ntsci")) {
    is_ntsc = 1;
    fwparams.EncodeParams.VFORMAT = 8;    
  }
  else {
    printf("Invalid norm!\n");
    return -1;
  }
  return 0;
}

int do_mon1(int argc, char *argv[]) {
  AP5200_MODE_PARAMS p;
  
  p.Mode = MODE_MON_FULL;
  p.MON_FULL.Channel = strtol(argv[1], NULL, 0);
  return simple_e0904_ioctl(enc, IOCTL_ETHJPG2K_VIDEO_SET_MODE, (void *) &p);
}

int do_pb(int argc, char *argv[]) {
  AP5200_PLAYBACK p;
  
  if (strtol(argv[1], NULL, 0) == -1) {
    p.enable = 0;
  }
  else {
    p.enable = 1;
    p.isNtsc = is_ntsc;
    p.dest_ch = strtol(argv[1], NULL, 0);
  }
  return simple_e0904_ioctl(enc, IOCLT_ETHJPG2K_VIDEO_PLAYBACK, (void *) &p);
}

int do_mon4(int argc, char *argv[]) {
  AP5200_MODE_PARAMS p;
  int i;
  
  p.Mode = MODE_MON_QUAD;
  for(i=0; i<4; i++)
    p.MON_QUAD.Channel[i] = strtol(argv[1+i], NULL, 0);
  return simple_e0904_ioctl(enc, IOCTL_ETHJPG2K_VIDEO_SET_MODE, (void *) &p);
}

int do_mon8(int argc, char *argv[]) {
  AP5200_MODE_PARAMS p;
  int i;
  
  p.Mode = MODE_MON_ALL;
  for(i=0; i<8; i++)
    p.MON_ALL.Channel[i] = strtol(argv[1+i], NULL, 0);
  return simple_e0904_ioctl(enc, IOCTL_ETHJPG2K_VIDEO_SET_MODE, (void *) &p);
}

int do_rec1(int argc, char *argv[]) {
  AP5200_MODE_PARAMS p;

  p.Mode = MODE_REC_FULL;
  p.REC_FULL.Channel = strtol(argv[1], NULL, 0);
  return simple_e0904_ioctl(enc, IOCTL_ETHJPG2K_VIDEO_SET_MODE, (void *) &p);
}

int do_rec4(int argc, char *argv[]) {
  AP5200_MODE_PARAMS p;
  int bSlave;

  if(strtol(argv[1], NULL, 0) == 1)
    bSlave = 1;
  else
    bSlave = 0;
  
  p.Mode = MODE_REC_QUAD;
  p.REC_QUAD.bSlave = bSlave;
  return simple_e0904_ioctl(enc, IOCTL_ETHJPG2K_VIDEO_SET_MODE, (void *) &p);
}

int do_saa(int code, char *argv[]) {
  unsigned char p[2];
  
  p[0] = strtol(argv[1], NULL, 0);
  p[1] = strtol(argv[2], NULL, 0);
  return simple_e0904_ioctl(enc, code, (void *) p);
}

int do_brightness(int argc, char *argv[]) {
  return do_saa(IOCTL_ETHJPG2K_VIDEO_SET_BRIGHTNESS, argv);
}

int do_contrast(int argc, char *argv[]) {
  return do_saa(IOCTL_ETHJPG2K_VIDEO_SET_CONTRAST, argv);
}

int do_hue(int argc, char *argv[]) {
  return do_saa(IOCTL_ETHJPG2K_VIDEO_SET_HUE, argv);
}

int do_clear(int argc, char *argv[]) {
  return simple_e0904_ioctl(enc, IOCLT_ETHJPG2K_VIDEO_CLEAR, NULL);
}

int do_osd_clear(int argc, char *argv[]) {
  int code;

  switch(strtol(argv[1], NULL, 0)) {
  case 1:
    code = IOCTL_ETHJPG2K_RECORD_MST_CLEAR_OSD;
    break;
  case 2:
    code = IOCTL_ETHJPG2K_RECORD_SLV_CLEAR_OSD;
    break;
  case 0:
  default:
    code = IOCTL_ETHJPG2K_MONITOR_CLEAR_OSD;
    break;
  }
  return simple_e0904_ioctl(enc, code, NULL);
}

int do_osd_write(int argc, char *argv[]) {
  int code;
  AP5200_OSD_PARAMS p;

  switch(strtol(argv[1], NULL, 0)) {
  case 1:
    code = IOCTL_ETHJPG2K_RECORD_MST_WRITE_OSD;
    p.bSlave = 0;
    break;
  case 2:
    code = IOCTL_ETHJPG2K_RECORD_SLV_WRITE_OSD;
    p.bSlave = 1;
    break;
  case 0:
  default:
    code = IOCTL_ETHJPG2K_MONITOR_WRITE_OSD;
    break;
  }
  p.x = strtol(argv[2], NULL, 0);
  p.y = strtol(argv[3], NULL, 0);
  p.String = argv[4];
  p.StrLen = strlen(argv[4]);
  p.Color = strtol(argv[5], NULL, 0);
  return simple_e0904_ioctl(enc, code, &p);
}

int do_slave_path(int argc, char *argv[]) {
  slave_path = strtol(argv[1], NULL, 0);
  return 0;
}

int do_init_board(int argc, char *argv[]) {
  if (running) {
    printf("Invalid while running\n");
    return 1;
  }

  assert(!simple_e0904_board_init(enc, is_ntsc, firmware, &fwdata, &fwparams, slave_path));

  return 0;
}

int do_noinit_board(int argc, char *argv[]) {
  if (running) {
    printf("Invalid while running\n");
    return 1;
  }

  simple_e0904_board_noinit(enc, is_ntsc);

  return 0;
}

int do_msleep(int argc, char *argv[]) {
  usleep(1000* strtol(argv[1], NULL, 0));
  return 0;
}

int do_motion_enable(int argc, char *argv[]) {
  AP5200_MOTION_ENABLE_CH e;
  int ch = strtol(argv[1], NULL, 0);
  
  e.ch = ch;
  e.show = 1;
  e.accumulate = 1;
  e.TH_L = 0x40;
  e.TH_H = 0xd0;
  e.TH_motion = 0x20;
  e.TH_blind = 1000;
  e.TH_win = 3;
  e.REF_catch = 50;
  e.sel_TH_motion = 0;
  memset(e.Windows, 0xff, sizeof(*(e.Windows)) * 12);
  
  return simple_e0904_ioctl(enc, IOCTL_ETHJPG2K_MOTION_ENABLE, (void *) &e);
}

int do_motion_enable_sf(int argc, char *argv[]) {
  AP5200_MOTION_ENABLE_CH e;
  int ch = strtol(argv[1], NULL, 0);
  
  e.ch = ch;
  e.show = strtol(argv[2], NULL, 0);
  e.accumulate = strtol(argv[3], NULL, 0);
  e.TH_L = strtol(argv[4], NULL, 0);
  e.TH_H = strtol(argv[5], NULL, 0);
  e.TH_motion = strtol(argv[6], NULL, 0);
  e.TH_blind = strtol(argv[7], NULL, 0);
  e.TH_win = strtol(argv[8], NULL, 0);
  e.REF_catch = strtol(argv[9], NULL, 0);
  e.sel_TH_motion = strtol(argv[10], NULL, 0);
  memset(e.Windows, 0xff, sizeof(*(e.Windows)) * 12);
  
  return simple_e0904_ioctl(enc, IOCTL_ETHJPG2K_MOTION_ENABLE, (void *) &e);
}

int do_motion_enable_f(int argc, char *argv[]) {
  AP5200_MOTION_ENABLE_CH e;
  int ch = strtol(argv[1], NULL, 0);
  int i;
  
  e.ch = ch;
  e.show = strtol(argv[2], NULL, 0);
  e.accumulate = strtol(argv[3], NULL, 0);
  e.TH_L = strtol(argv[4], NULL, 0);
  e.TH_H = strtol(argv[5], NULL, 0);
  e.TH_motion = strtol(argv[6], NULL, 0);
  e.TH_blind = strtol(argv[7], NULL, 0);
  e.TH_win = strtol(argv[8], NULL, 0);
  e.REF_catch = strtol(argv[9], NULL, 0);
  e.sel_TH_motion = strtol(argv[10], NULL, 0);
  for(i=0; i<12; i++)
    e.Windows[i] = strtol(argv[11+i], NULL, 0);
  
  return simple_e0904_ioctl(enc, IOCTL_ETHJPG2K_MOTION_ENABLE, (void *) &e);
}

int do_motion_disable(int argc, char *argv[]) {
  AP5200_MOTION_ENABLE_CH e;
  int ch = strtol(argv[1], NULL, 0);
  
  e.ch = ch;
  
  return simple_e0904_ioctl(enc, IOCTL_ETHJPG2K_MOTION_DISABLE, (void *) &e);
}

int do_motion_status(int argc, char *argv[]) {
  AP5200_MOTION_STATUS e;
  int i,j;
  
  simple_e0904_ioctl(enc, IOCTL_ETHJPG2K_MOTION_GET_STATUS, (void *) &e);
  for(i=0; i<8; i++) {
    if (e.ch[i].on) {
      printf("ch %d b %d m %d sigma %d w", i,
	     e.ch[i].BlindDetected,
	     e.ch[i].MotionDetected,
	     e.ch[i].sigma);
      for(j=0; j<12; j++)
	printf(" %04x", e.ch[i].Windows[j]);
      printf("\n");
    }
  }
  return 0;
}

int do_motion_status_fast(int argc, char *argv[]) {
  AP5200_MOTION_STATUS e;
  int i;
  
  simple_e0904_ioctl(enc, IOCTL_ETHJPG2K_MOTION_GET_STATUS_FAST, (void *) &e);
  for(i=0; i<8; i++) {
    if (e.ch[i].on) {
      printf("ch %d b %d m %d\n", i,
	     e.ch[i].BlindDetected,
	     e.ch[i].MotionDetected);
    }
  }
  return 0;
}



/* command table */
struct cmd_table_s {
  char *name;
  cmd_cb f;
  int argc;
  char *help;
  int need_restart;
} cmd_table[] = {
  { "every", do_every, 2, "takes a shot every [arg1] frames and saves to a file with prefix [arg2]" },
  { "fw", do_fw, 1, "pathname of firmware to use", 1},
  { "help", do_help, 0, "this help" },
  { "shot", do_shot, 1, "takes a shot of a picture to filename [arg1]" },
  { "stat", do_stat, 0, "display statistics: running status, frames got, dropped, resets" },
  { "start", do_start, 0, "starts encoding" },
  { "stop", do_stop, 0, "stops encoding" },
  { "quit", do_quit, 0, "exits this application" },
  { "verbose", do_verbose, 1, "if [arg1] is 0 be quiet, if 1 show just errors, if 2 be verbose" },
  { "wait", do_wait, 1, "wait [arg1] frames before continuing" },
  { "quality", do_quality, 2, "applies scheme [arg1] to quality of generated images.\n[arg1] = 0 no truncation\n[arg1] = 1 images are +/- 5% of dimensions specified by [arg2]\n[arg1] = 2 [arg2] is the quality factor (best = 0x100, very bad = 0x900)", 1},
  { "norm", do_norm, 1, "[arg1] can be pal, ntsc, pali, ntsci", 1},
  { "mon1", do_mon1, 1, "monitor path full channel [arg1]"},
  { "mon4", do_mon4, 4, "monitor path 4 channel [arg1]-[arg4]"},
  { "mon8", do_mon8, 8, "monitor path 8 channel [arg1]-[arg8]"},
  { "rec1", do_rec1, 1, "record path to full channel [arg1] is 0-7"},
  { "rec4", do_rec4, 1, "record path for master([arg1] = 0) or slave([arg1] = 1) to quad"},
  { "brightness", do_brightness, 2, "set brightness of channel [arg1] to [arg2]"},
  { "contrast", do_contrast, 2, "set contrast of channel [arg1] to [arg2]"},
  { "hue", do_hue, 2, "set hue of channel [arg1] to [arg2]"},
  { "clear", do_clear, 0, "clear AP5200 video memory"},
  { "osd_clear", do_osd_clear, 1, "clear OSD, [arg1] is 0 for monitor, 1 for record master, 2 for record slave"},
  { "osd_write", do_osd_write, 5, "write OSD, [arg1] is 0 for monitor, 1 for record master, 2 for record slave\n[arg2],[arg3] are x,y coordinates\n[arg4] is the string to write *WITHOUT* spaces\n[arg5] is the color, 0 default, 0x100 yellow, 0x200 blue, 0x300 red"},
  { "slave_path", do_slave_path, 1, "[arg1] = 1 sets the record path for the slave AP52000"},
  { "init_board", do_init_board, 0, "Initializes companion chips to the ADV202"},
  { "noinit_board", do_noinit_board, 0, "Simulates the initializaion of the board. Useful is already initialized."},
  { "msleep", do_msleep, 1, "sleeps [arg1] ms"},
  { "skip", do_skip, 1, "skips [arg1] frames between 2 acquired pictures in hardware. [arg1] is in the range 0-63."},
  { "motion_enable", do_motion_enable, 1, "enables motion detection on channel [arg1]"},
  { "motion_enable_sf", do_motion_enable_sf, 10, "enables motion detection on channel. args: [ch] [show] [accumulate] [TH_L] [TH_H] [TH_motion] [TH_blind] [TH_win] [REF_catch] [sel_TH_motion]"},
  { "motion_enable_f", do_motion_enable_f, 10+12, "enables motion detection on channel. args: [ch] [show] [accumulate] [TH_L] [TH_H] [TH_motion] [TH_blind] [TH_win] [REF_catch] [sel_TH_motion] [win 1-12]"},
  { "motion_disable", do_motion_disable, 1, "disables motion detection on channel [arg1]"},
  { "motion_status", do_motion_status, 0, "motion status"},
  { "motion_status_fast", do_motion_status_fast, 0, "motion status (fast and incomplete version)"},
  { "playback", do_pb, 1, " active playback on channel [arg1]. arg == -1 disable"},
  { NULL, },
};

/* help needs to have the table of command */
int do_help(int argc, char *argv[]) {
  struct cmd_table_s *c = cmd_table;
  int i;
  
  printf("Available commands: (*) means must restart encoding to take effects\n");
  while(c->name) {
    printf("-> %s", c->name);
    for(i=1; i < c->argc+1; i++) {
      printf(" [arg%d]", i);
    }
    printf("%s\n", c->need_restart ? " (*)" : "");
    printf("%s\n", c->help);
    c++;
  }
  return 0;
}

/* command parser */
int parse_cmd(char *c) {
#define MAX_ARGV 30
  char *argv[MAX_ARGV];
  int i, ret = 0;
  struct cmd_table_s *cmd = cmd_table;

  argv[0] = c;
  i = 1;
  while(*c) {
    if(*c != ' ')
      c++;
    else {
      *c = '\0';
      c++;
      while (*c && *c == ' ')
	c++;
      if (*c == '\0')
	break;
      argv[i] = c;
      i++;
    }
  }
  while(cmd->name) {
    if (!strcmp(cmd->name, argv[0])) {
      if (cmd->argc != i-1) {
	printf("Bad number of args, got %d expected %d\n", i, cmd->argc);
	ret = -1;
	break;
      }
      ret = cmd->f(i, argv);
      break;
    }
    cmd++;
  }
  if (!cmd->name) {
    printf("Unknown command\n");
    ret = -2;
  }
  if (ret == 0 || ret == -666)
    printf("OK\n");
  else
    printf("ERR %d\n", ret);
  return ret;
}



int main(int argc, char *argv[]) {
  char line[200];
  int i, ret = 0, len;

  if (argc < 2) {
    printf("Usage: %s [dev] [commands...]\n Command are also read from stdinput.\n For a list of command: %s [dev] help quit\n",
	   argv[0], argv[0]);
  }

  enc = simple_e0904_encoder(argv[1]);
  assert(enc);
  
  for(i=2; i<argc; i++) {
    ret = parse_cmd(argv[i]);
    if (ret == -666)
      return 0;
  }

  while (fgets(line, 200, stdin) != NULL) {
    len = strlen(line);
    if(line[len-1] == '\n')
      line[len-1] = '\0';
    ret = parse_cmd(line);
    if (ret == -666)
      return 0;
  }

  return 0;
}
